import logging
from dateutil.parser import parse
import os
import time

import pyspedas
from pyspedas.tplot_tools import options, link, cdf_to_tplot, sts_to_tplot

from .download_files_utilities import (
    set_new_data_root_dir,
    get_root_data_dir,
    create_dir_if_needed,
    get_orbit_files,
    get_filenames,
    get_new_files,
    get_file_from_site,
    display_progress,
)
from .file_regex import maven_kp_l2_regex  # kp_regex, l2_regex
from .orbit_time import orbit_time
from .maven_kp_to_tplot import maven_kp_to_tplot

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def maven_filenames(
    filenames=None,
    instruments=None,
    level="l2",
    insitu=True,
    iuvs=False,
    start_date="2014-01-01",
    end_date="2014-01-02",
    update_prefs=False,
    only_update_prefs=False,
    local_dir=None,
    public=True,
):
    """
    This function queries the MAVEN SDC API, and will return a list of files that match the given parameters.

    Parameters
    ----------
    filenames : list of str, optional
        Predefined list of filenames. If provided, other parameters are ignored. Defaults to None.
    instruments : list of str, optional
        List of instrument names. Defaults to None.
    level : str, optional
        Data level. Defaults to 'l2'.
    insitu : bool, optional
        If True, insitu files will be used. Defaults to True.
    iuvs : bool, optional
        If True, iuvs files will be used. Defaults to False.
        When True, file_extension=tab is added to the query.
    start_date : str/int, optional
        Start date in 'YYYY-MM-DD' format, or the orbit number. Defaults to '2014-01-01'.
    end_date : str/int, optional
        End date in 'YYYY-MM-DD' format, or the orbit number. Defaults to '2014-01-02'.
    update_prefs : bool, optional
        If True, updates preferences. Defaults to False.
    only_update_prefs : bool, optional
        If True, only updates preferences and does not return filenames. Defaults to False.
    local_dir : str, optional
        Local directory to use. Defaults to None.
    public: bool, optional
        If False, try loading data from the non-public service


    Returns
    -------
    list of str
        List of MAVEN filenames.
    """

    # Check for orbit num rather than time string
    if isinstance(start_date, int) and isinstance(end_date, int):
        logging.info(
            "Orbit numbers specified, checking for updated orbit # file from naif.jpl.nasa.gov"
        )
        get_orbit_files()
        start_date, end_date = orbit_time(start_date, end_date)
        start_date = parse(start_date)
        end_date = parse(end_date)
        start_date = start_date.replace(hour=0, minute=0, second=0)
        end_date = end_date.replace(day=end_date.day + 1, hour=0, minute=0, second=0)
        start_date = start_date.strftime("%Y-%m-%d")
        end_date = end_date.strftime("%Y-%m-%d")

    # Update the preference directory
    if update_prefs or only_update_prefs:
        set_new_data_root_dir()
        if only_update_prefs:
            return

    # Check for public vs private access
    # Hard code in access as public for now
    # public = True
    """
    public = get_access()
    if not public:
        get_uname_and_password()
    """
    # If no instruments are specified, default to the KP data set
    if instruments is None:
        instruments = ["kp"]
        if insitu:
            level = "insitu"
        if iuvs:
            level = "iuvs"

    if level == "kp":
        instruments = ["kp"]
        if insitu:
            level = "insitu"
        if iuvs:
            level = "iuvs"

    # Set data download location
    if local_dir is None:
        mvn_root_data_dir = get_root_data_dir()
    else:
        mvn_root_data_dir = local_dir

    # Keep track of files to download
    maven_files = {}
    for instrument in instruments:
        # Build the query to the website
        query_args = []
        query_args.append("instrument=" + instrument)
        query_args.append("level=" + level)
        if filenames is not None:
            query_args.append("file=" + filenames)
        query_args.append("start_date=" + start_date)
        query_args.append("end_date=" + end_date)
        if level == "iuvs":
            query_args.append("file_extension=tab")

        sep = "/" if is_fsspec_uri(mvn_root_data_dir) else os.path.sep
        data_dir = sep.join([
            mvn_root_data_dir, "maven", "data", "sci", instrument, level
        ])

        query = "&".join(query_args)

        logging.info("Querying for filenames: %s", query)
        s = get_filenames(query, public)

        if not s:
            logging.error("No files found for {}.".format(instrument))
            if instrument not in maven_files:
                maven_files[instrument] = []
            continue

        s = s.split(",")
        logging.info("Returned %d files",len(s))

        if instrument not in maven_files:
            maven_files[instrument] = [s, data_dir, public]
        else:
            maven_files[instrument].extend([s, data_dir, public])

    # Grab KP data too, there is a lot of good ancillary info in here
    if instruments != "kp":
        instrument = "kp"
        # Build the query to the website
        query_args = []
        query_args.append("instrument=kp")
        query_args.append("level=insitu")
        query_args.append("start_date=" + start_date)
        query_args.append("end_date=" + end_date)
        sep = "/" if is_fsspec_uri(mvn_root_data_dir) else os.path.sep
        data_dir = sep.join([
            mvn_root_data_dir, "maven", "data", "sci", "kp", "insitu"
        ])
        query = "&".join(query_args)
        s = get_filenames(query, public)
        if not s:
            logging.error("No files found for {}.".format(instrument))
            if instrument not in maven_files:
                maven_files[instrument] = []
        else:
            s = s.split(",")
            if instrument not in maven_files:
                maven_files[instrument] = [s, data_dir, public]
            else:
                maven_files[instrument].extend([s, data_dir, public])

    return maven_files


def maven_file_groups(files):
    """
    This function groups MAVEN files by their description.

    Parameters
    ----------
    files : list of str
        List of MAVEN filenames.

    Returns
    -------
    dict
        Dictionary of grouped files.
    """

    # Group files by their description
    result = {}
    if len(files) == 0:
        return result
    files.sort()

    kp_regex, l2_regex = maven_kp_l2_regex()
    for f in files:
        if is_fsspec_uri(f):
            protocol, path = f.split("://")
            fs = fsspec.filesystem(protocol)

            basename = f.rstrip("/").split("/")[-1]
        else:
            basename = os.path.basename(f)
        desc = l2_regex.match(basename).group("description")
        if desc not in result:
            result[desc] = []
        result[desc].append(f)

    for k in result.keys():
        result[k].sort()

    return result


def load_data(
    filenames=None,
    instruments=None,
    level="l2",
    type=None,
    insitu=True,
    iuvs=False,
    start_date="2014-01-01",
    end_date="2014-01-02",
    update_prefs=False,
    only_update_prefs=False,
    local_dir=None,
    list_files=False,
    new_files=True,
    exclude_orbit_file=False,
    download_only=False,
    varformat=None,
    varnames=[],
    prefix='',
    suffix='',
    get_support_data=False,
    get_metadata=False,
    auto_yes=False,
    public=True,
):
    """
    This function downloads MAVEN data loads it into tplot variables, if applicable.

    Parameters
    ----------
    filenames : list of str, optional
        Predefined list of filenames. If provided, other parameters are ignored. Defaults to None.
    instruments : list of str, optional
        List of instrument names. Defaults to None.
        Accepted values are any combination of: sta, swi, swe, lpw, euv, ngi, iuv, mag, sep, rse
    level : str, optional
        Data level. Defaults to 'l2'. Currently unused, only l2 data can be loaded.
    type : str, optional
        Type of data to load. Defaults to None.

        The observation/file type of the instruments to load.  If None, all file types are loaded.
        Otherwise, a file will only be loaded into tplot if its descriptor matches one of the strings in this field.

        Accepted values are:
        =================== ====================================
        Instrument           Level 2 Observation Type/File Type
        =================== ====================================
        EUV                 bands
        LPW                 lpiv, lpnt, mrgscpot, we12, we12burstlf, we12bursthf, we12burstmf, wn, wspecact, wspecpas
        STATIC              2a, c0, c2, c4, c6, c8, ca, cc, cd, ce, cf, d0, d1, d4, d6, d7, d8, d9, da, db
        SEP                 s1-raw-svy-full, s1-cal-svy-full, s2-raw-svy-full, s2-cal-svy-full
        SWEA                arc3d, arcpad, svy3d, svypad, svyspec
        SWIA                coarsearc3d, coarsesvy3d, finearc3d, finesvy3d, onboardsvymom, onboardsvyspec
        MAG                 ss, pc, pl, ss1s, pc1s, pl1s
        =================== =====================================
    insitu : bool, optional
        If True, only insitu files will be used. Defaults to True.
    iuvs : bool, optional
        If True, iuvs files will be used. Defaults to False.
        When True, file_extension=tab is added to the query.
    start_date : str/int, optional
        Start date in 'YYYY-MM-DD' format, or the orbit number. Defaults to '2014-01-01'.
    end_date : str/int, optional
        End date in 'YYYY-MM-DD' format, or the orbit number. Defaults to '2014-01-02'.
    update_prefs : bool, optional
        If True, updates preferences. Defaults to False.
    only_update_prefs : bool, optional
        If True, only updates preferences and does not return data. Defaults to False.
    local_dir : str, optional
        Local directory to use. Defaults to None.
    list_files : bool, optional
        If True, lists files without loading data. Defaults to False.
    new_files : bool, optional
        If True, get new files. Defaults to True.
    exclude_orbit_file : bool, optional
        If True, will not download the latest orbit table. Defaults to False.
        Currently unused.
    download_only : bool, optional
        If True, only downloads files without loading data. Defaults to False.
    varformat : str, optional
        Variable format. Defaults to None.
        The file variable formats to load into tplot. Wildcard character
        "*" is accepted.  By default, all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load. Defaults to [].
    prefix : str, optional
        Prefix to append to variable names. Defaults to ''.
    suffix : str, optional
        Suffix to append to variable names.
        Defaults to ''.
        By default the part of the filename (plus the suffix) is appended to the variable name.
        If suffix='empty', no suffix is appended to the variable name (no part of the filename and no suffix).
    get_support_data : bool, optional
        If True, retrieves support data. Defaults to False.
    get_metadata : bool, optional
        If True, retrieves metadata. Defaults to False.
    auto_yes : bool, optional
        If True, automatically answers 'yes' to prompts. Defaults to False.
    public: bool, optional
    If false, try using the non-public interface


    Returns
    -------
    dict
        Dictionary of loaded data variables.
    """

    if not isinstance(instruments, list) and instruments is not None:
        instruments = [instruments]

    if not isinstance(type, list) and type is not None:
        type = [type]

    if not isinstance(filenames, list) and filenames is not None:
        filenames = [filenames]

    kp_regex, l2_regex = maven_kp_l2_regex()

    # 1. Get a list of MAVEN files queries from the above seach parameters
    maven_files = maven_filenames(
        filenames,
        instruments,
        level,
        insitu,
        iuvs,
        start_date,
        end_date,
        update_prefs,
        only_update_prefs,
        local_dir,
        public=public,
    )

    # If we are not asking for KP data, this flag ensures only ancillary data is loaded in from the KP files
    if level != "kp":
        ancillary_only = True
    else:
        ancillary_only = False

    # Convert to list
    if not isinstance(type, list):
        type = [type]

    # Keep track of what files are downloaded
    files_to_load = []

    # Loop through all instruments, download files locally if needed
    for instr in maven_files.keys():
        bn_files_to_load = []
        # There might be more than one set of files to load for this instrument (e.g. kp with iuvs)
        maven_file_list = maven_files[instr]
        while len(maven_file_list) > 0:
            s = maven_file_list[0]
            data_dir = maven_file_list[1]
            public = maven_file_list[2]
            # Strip off these three entries, in case there's more to process next iteration
            maven_file_list = maven_file_list[3:]

            # Add to list of files to load
            for f in s:
                # Filter by type
                if type != [None] and instr != "kp":
                    file_type_match = False
                    desc = l2_regex.match(f).group("description")
                    for t in type:
                        # kluge for STATIC, jmm, 2024-07-24, otherwise type='d1' results in ca,d0,d1,d4 loading
                        if instr == "sta":
                            if t + "-" in desc:
                                file_type_match = True
                        else:
                            if t in desc:
                                file_type_match = True
                    if not file_type_match:
                        continue

                # Check if the files are KP data
                try:
                    if instr == "kp":
                        full_path = create_dir_if_needed(f, data_dir, "insitu")
                    else:
                        full_path = create_dir_if_needed(f, data_dir, level)
                    bn_files_to_load.append(f)
                    sep = "/" if is_fsspec_uri(full_path) else os.path.sep
                    files_to_load.append(sep.join([full_path, f]))
                except Exception as e:
                    # todo: better handling of rse .tab files
                    # rse files are .tab files (TAB delimited text files) that currently cannot be loaded into tplot
                    # for now, we need to skip these files
                    logging.error("Cannot handle file: " + f)
                    continue

            if list_files:
                for f in s:
                    logging.info(f)
                return

            if new_files:
                if instr == "kp":
                    s = get_new_files(bn_files_to_load, data_dir, instr, "insitu")
                else:
                    s = get_new_files(bn_files_to_load, data_dir, instr, level)
            if len(s) == 0:
                continue
            logging.info(
                "Your request will download a total of: "
                + str(len(s))
                + " files for instrument "
                + str(instr)
            )
            logging.info("Files: ")
            logging.info(s)
            logging.info("Would you like to proceed with the download? ")
            valid_response = False
            cancel = False
            if auto_yes:
                valid_response = True
                logging.info("- Auto yes")
            while not valid_response:
                response = input("(y/n) >  ")
                if response == "y" or response == "Y":
                    valid_response = True
                    cancel = False
                elif response == "n" or response == "N":
                    logging.error("Cancelled download. Returning...")
                    valid_response = True
                    cancel = True
                else:
                    logging.error("Invalid input.  Please answer with y or n.")

            if cancel:
                continue

            i = 0
            display_progress(i, len(s))
            for f in s:
                logging.info("File:" + str(f))
                i = i + 1
                if instr == "kp":
                    full_path = create_dir_if_needed(f, data_dir, "insitu")
                else:
                    full_path = create_dir_if_needed(f, data_dir, level)
                get_file_from_site(f, public, full_path)
                time.sleep(10.0)
                display_progress(i, len(s))

    # 2. Load files into tplot

    if files_to_load:
        # Flatten out downloaded files from list of lists of filenames
        if isinstance(files_to_load[0], list):
            files_to_load = [item for sublist in files_to_load for item in sublist]

        # Only load in files into tplot if we actually downloaded CDF files
        cdf_files = [f for f in files_to_load if ".cdf" in f]
        sts_files = [f for f in files_to_load if ".sts" in f]
        kp_files = [f for f in files_to_load if ".tab" in f]

        loaded_tplot_vars = []
        if not download_only:

            # Load in CDF files
            if len(cdf_files) > 0:
                cdf_dict = maven_file_groups(cdf_files)
                for desc in cdf_dict.keys():
                    if suffix == "empty":
                        # In this case, no suffix is appended to the variable name
                        suf = ""
                    else:
                        # The description (part of the filename) is appended to the variable name
                        suf = desc + suffix
                    created_vars = cdf_to_tplot(
                        cdf_dict[desc],
                        varformat=varformat,
                        varnames=varnames,
                        string_encoding="utf-8",
                        get_support_data=get_support_data,
                        prefix=prefix,
                        suffix=suf,
                        get_metadata=get_metadata,
                    )
                    # Specifically for SWIA and SWEA data, make sure the plots have log axes and are spectrograms
                    if is_fsspec_uri(cdf_dict[desc][0]):
                        protocol, path = cdf_dict[desc][0].split("://")
                        fs = fsspec.filesystem(protocol)

                        basename = cdf_dict[desc][0].rstrip("/").split("/")[-1]
                    else:
                        basename = os.path.basename(cdf_dict[desc][0])
                    instr = l2_regex.match(basename).group(
                        "instrument"
                    )
                    if instr in ["swi", "swe"]:
                        options(created_vars, "spec", 1)
                    loaded_tplot_vars.append(created_vars)

            # Load in STS files
            if len(sts_files) > 0:
                sts_dict = maven_file_groups(sts_files)
                for desc in sts_dict.keys():
                    if suffix == "empty":
                        # In this case, no suffix is appended to the variable name
                        suf = ""
                    else:
                        # The description (part of the filename) is appended to the variable name
                        suf = desc + suffix
                    try:
                        created_vars = sts_to_tplot(
                            sts_dict[desc],
                            prefix=prefix,
                            suffix=suf,
                        )
                    except FileNotFoundError:
                        logging.error("PyTplot Error: STS importer is not URI capable.")
                        logging.error("\tSkipping file as PyTplot cannot use this type of filesystem.")
                        continue
                    loaded_tplot_vars.append(created_vars)

                # Remove the Decimal Day column, not really useful
                for tvar in loaded_tplot_vars:
                    if "DDAY_" in tvar:
                        pyspedas.tplot_tools.del_data(tvar)
                        del tvar

            # Flatten out the list and only grab the unique tplot variables
            flat_list = list(
                set([item for sublist in loaded_tplot_vars for item in sublist])
            )

            # Load in KP data specifically for all of the Ancillary data (position, attitude, Ls, etc)
            if kp_files != []:
                kp_data_loaded = maven_kp_to_tplot(
                    filename=kp_files,
                    ancillary_only=ancillary_only,
                    instruments=instruments,
                )

                logging.info("Kp data loaded:")
                logging.info(kp_data_loaded)
                # Link all created KP data to the ancillary KP data
                for tvar in kp_data_loaded:
                    link(tvar, "mvn_kp::spacecraft::altitude", link_type="alt")
                    link(tvar, "mvn_kp::spacecraft::mso_x", link_type="x")
                    link(tvar, "mvn_kp::spacecraft::mso_y", link_type="y")
                    link(tvar, "mvn_kp::spacecraft::mso_z", link_type="z")
                    link(tvar, "mvn_kp::spacecraft::geo_x", link_type="geo_x")
                    link(tvar, "mvn_kp::spacecraft::geo_y", link_type="geo_y")
                    link(tvar, "mvn_kp::spacecraft::geo_z", link_type="geo_z")
                    link(
                        tvar, "mvn_kp::spacecraft::sub_sc_longitude", link_type="lon"
                    )
                    link(
                        tvar, "mvn_kp::spacecraft::sub_sc_latitude", link_type="lat"
                    )

            # Link all created tplot variables to the corresponding KP data
            for tvar in flat_list:
                link(tvar, "mvn_kp::spacecraft::altitude", link_type="alt")
                link(tvar, "mvn_kp::spacecraft::mso_x", link_type="x")
                link(tvar, "mvn_kp::spacecraft::mso_y", link_type="y")
                link(tvar, "mvn_kp::spacecraft::mso_z", link_type="z")
                link(tvar, "mvn_kp::spacecraft::geo_x", link_type="geo_x")
                link(tvar, "mvn_kp::spacecraft::geo_y", link_type="geo_y")
                link(tvar, "mvn_kp::spacecraft::geo_z", link_type="geo_z")
                link(
                    tvar, "mvn_kp::spacecraft::sub_sc_longitude", link_type="lon"
                )
                link(
                    tvar, "mvn_kp::spacecraft::sub_sc_latitude", link_type="lat"
                )

            logging.info("Maven data list:")
            logging.info(flat_list)

            # Return list of unique KP data
            return flat_list



import numpy as np


def mpause_2(xmp=None, ymp_west=None, short=False, xmp_max=10.78):
    """
    Calculate the magnetopause (X, Y) locations based on the Fairfield model (JGR, 1971).
    An aberration of 4 degrees is assumed.

    Parameters
    ----------
    xmp (array-like, optional):
        Spacecraft position, x component (defaults to a range if not provided).
    ymp_west (deprecated, unused):
        This parameter is kept for compatibility with the IDL version, but it is unused.
    short (bool, optional):
        If True, the function returns a shorter version of the magnetopause (defaults to False).
    xmp_max (float, optional):
        Maximum value of xmp (defaults to 10.78, the value from the Fairfield paper).

    Returns
    -------

    tuple: A tuple containing two elements:
        xmp (numpy.ndarray): The x coordinates of the magnetopause.
        ymp (numpy.ndarray): The y coordinates of the magnetopause.


    Notes
    -----

    Similar to mpause_2.pro in IDL SPEDAS.

    """

    # Define constants
    xp = -15.
    xmp_min = -300
    npoints = 1000
    aberangle = 4.5  # This constant is defined but not used in this implementation

    # Default xmp values if not provided
    if xmp is None:
        xmp = np.linspace(xmp_min, xmp_max, npoints)

    # Ellipse coefficients from the Fairfield model
    a1 = 0.0278
    b1 = 0.3531
    c1 = -0.586
    d1 = 17.866
    e1 = -233.67

    # Split the input xmp array into two parts based on the threshold xp
    ilt15 = xmp < xp
    ige15 = xmp >= xp

    ymp_east = np.copy(xmp)  # Initialize ymp_east with the xmp values
    ymp_west = np.copy(xmp)  # Initialize ymp_west with the xmp values

    # Calculate for xmp >= -15
    if np.any(ige15):
        beta = a1 * xmp[ige15] + c1
        gamma = b1 * xmp[ige15]**2 + d1 * xmp[ige15] + e1
        delta = beta**2 - 4 * gamma

        # Compute ymp_east and ymp_west using the quadratic formula
        ymp_east[ige15] = (-beta - np.sqrt(delta)) / 2.
        ymp_west[ige15] = (-beta + np.sqrt(delta)) / 2.

        # Calculating the slopes for interpolation
        in15 = np.argmin(xmp[ige15])
        ww = -1 if in15 == len(xmp[ige15]) - 1 else 1
        s_east = (ymp_east[ige15][in15 + ww] - ymp_east[ige15]
                  [in15]) / (xmp[ige15][in15 + ww] - xmp[ige15][in15])
        s_west = (ymp_west[ige15][in15 + ww] - ymp_west[ige15]
                  [in15]) / (xmp[ige15][in15 + ww] - xmp[ige15][in15])

        # Interpolate for xmp < -15
        if np.any(ilt15):
            ymp_east[ilt15] = ymp_east[ige15][in15] + \
                s_east * (xmp[ilt15] - xmp[ige15][in15])
            ymp_west[ilt15] = ymp_west[ige15][in15] + \
                s_west * (xmp[ilt15] - xmp[ige15][in15])

    # Construct the output arrays depending on the 'short' parameter
    if not short:
        # Indices of the array in reverse order
        ireverse = np.arange(len(xmp))[::-1]
        # Duplicate xmp with reversed order for symmetry
        xmp = np.concatenate((xmp, xmp[ireverse]))
        # Combine ymp_east and reversed ymp_west
        ymp_east = np.concatenate((ymp_east, ymp_west[ireverse]))
        ymp = ymp_east
    else:
        ymp = ymp_east

    return xmp, ymp



from functools import update_wrapper
from pyspedas.utilities.pyspedas_functools import better_partial
from .load import load
from pyspedas.utilities.datasets import find_datasets


# Define partial wrappers for other load routines, fixing the instrument parameter
# 'better_partial' works better with PyCharm autocompletion than functools.partial
# update_wrapper() is necessary for help() to show info for the wrapped function,
# rather than the partial() object.

mam = better_partial(load, instrument="mam")
update_wrapper(mam, load)
edi = better_partial(load, instrument="edi")
update_wrapper(edi, load)
epi = better_partial(load, instrument="epi")
update_wrapper(epi, load)
ici = better_partial(load, instrument="ici")
update_wrapper(ici, load)
pcd = better_partial(load, instrument="pcd")
update_wrapper(pcd, load)
sfd = better_partial(load, instrument="sfd")
update_wrapper(sfd, load)

datasets = better_partial(find_datasets, mission="Equator-S", label=True)
update_wrapper(datasets, find_datasets)



import numpy as np
from pyspedas import tnames,data_exists
from pyspedas import tinterpol


def kp2iopt(kp, varname=None, plus1=False):
    """
    Convert a tplot or array Kp index variable to iopt values suitable for passing to the T89 field
    modeling or tracing routines.

    Input Kp values will be in the range of 0 to 6, and may have a fractional part indicating an increasing
    or decreasing index. For example, 2.33333 represents 2+, 2.66666 represents 3-. iopt output values
    will be in the range 1 to 7.

    Parameters:
    ----------
    kp : scalar or array-like
        Kp values (e.g. from noaa_load_kp) or name of a tplot variable giving the Kp index values.

    varname : str, optional
        A string specifying a tplot position variable. Required if kp is a tplot variable.
        Output iopt values will be interpolated to the 'varname' times using nearest-neighbor interpolation.

    plus1 : bool, optional
        If specified, overrides the default behavior of rounding iopt to the nearest integer, and
        instead sets iopt to kp+1, preserving any fractional part.

    Returns:
    -------
    array-like
        IOPT values suitable for passing to the GEOPACK T89 tracing and modeling routines. Represented
        as double precision floating point for compatibility with GEOPACK library calling sequences,
        but may or may not be rounded to integer values depending on whether the kp_plus1 keyword is specified.
    """

    # Check if kp is a string
    if isinstance(kp, str):
        # Assume kp is stored in a tplot variable
        if not data_exists(kp):
            raise ValueError(f'kp is of type string but no tplot variable exists with name={kp}')

        # Interpolate kp values
        tinterpol(kp, varname, newname='kp_int_tmp',method="nearest")
        kp_dat = kp

    # Convert kp_dat to iopt values
    if plus1:
        iopt = kp_dat + 1.0
    else:
        iopt = np.floor(kp_dat + 1.5)

    # Force to valid iopt range
    iopt = np.clip(iopt, 1.0, 7.0)

    return iopt


import numpy as np
from pyspedas.tplot_tools import get_data, store_data


def tnormalize(variable, newname=None, return_data=False):
    """
    Normalize all the vectors stored in a tplot variable

    Parameters
    ----------
        variable: str or np.ndarray
            tplot variable (or numpy array) containing the vectors to be normalized

        newname: str, optional
            name of the output variable; If not set, output name will be variable + '_normalize'
            Default: None

        return_data: bool, optional
            If true, return the normalized vectors instead of creating a tplot variable
            Default: False
    
    Returns
    -------
        str or ndarray of float
            name of the tplot variable created or normalized vectors if return_data is set
    
    Examples
    ---------
        >>> import pyspedas
        >>> pyspedas.store_data('a', data={'x':[1, 2, 3], 'y':[[1,2,3],[4,5,6],[7,8,9]]})
        >>> pyspedas.tnormalize('a')


    """
    metadata_in = {}
    if isinstance(variable, str):
        data_in = get_data(variable)
        metadata_in = get_data(variable, metadata=True)
        data = data_in[1]
        times = data_in[0]
    else:
        data = np.atleast_2d(variable)
        times = np.zeros(data.shape[0])

    n = np.sqrt(np.nansum(data**2, axis=1))

    # to do element-wise division, the magnitude needs to be repeated for each component
    norm_reshaped = np.reshape(n, [len(times), 1])
    norm_mag = np.repeat(norm_reshaped, len(data[0, :]), axis=1)

    data_norm = data/norm_mag

    if return_data:
        return data_norm
    else:
        if newname is None:
            newname = variable + '_normalized'
        store_data(newname, data={'x': times, 'y': data_norm}, attr_dict=metadata_in)
        return newname



import pyspedas
from pyspedas.tplot_tools import tplot_wildcard_expand
import logging

def annotate(tvar=None, text=None, position=None,
             xycoords='axes fraction',
             color='black',
             fontfamily=None,
             fontsize='x-large',
             alpha=1,
             fontvariant='normal',
             fontstyle='normal',
             fontstretch='normal',
             fontweight='normal',
             rotation='horizontal',
             delete=False):
    """
    Add text annotatons to tplot panels.

    Most of the parameters are passed to the matplotlib axes.annotate() method as a dictionary.

    Parameters
    ==========

    tvar: str
        A tplot variable name or list of names to add annotations to (wildcards accepted).
    text: str
        Text to add to the plot
    position: tuple or array
        The location for the annotation to appear
    xycoords: str
        The coordinate system to use for the position. 'axes fraction' interprets the positions relative to the panel size. 'data' uses
        data coordinates (times for the x axis, data values for the y axis).  Times should be passed as np.datetime64 objects.
    color: str
        Color of the text (default: 'black')
    fontfamily: str
        Font to be used for the text
    fontsize: str or numeric
        Font size to use.  If numeric, units are in points. (Default: 'x-large')
    alpha: float
        Transparency value to use for the text (default: 1)
    fontvariant: str
        Other attributes of the font to use
    fontstyle: str
        Other attributes of the font to use
    fontstretch: str
        Other attributes of the font to use
    fontweight: str
        Other attributes of the font to use
    rotation:str
        Text orientation ('horizontal', 'vertical')
    delete: bool
        If True, delete all annotations for this variable.

    """

    if isinstance(tvar, int):
        var = list(pyspedas.tplot_tools.data_quants.keys())[tvar]

    if not isinstance(tvar, list):
        tvar = [tvar]

    names = tplot_wildcard_expand(tvar)
    if len(names) == 0:
        logging.warning("annotations: no valid tplot variables specified")
        return

    annotations = {'text': text,
                   'position': position,
                   'xycoords': xycoords,
                   'fontfamily': fontfamily,
                   'fontsize': fontsize,
                   'fontvariant': fontvariant,
                   'fontstyle': fontstyle,
                   'fontstretch': fontstretch,
                   'fontweight': fontweight,
                   'rotation': rotation,
                   'color': color,
                   'alpha': alpha}

    for name in names:

        if name not in pyspedas.tplot_tools.data_quants.keys():
                logging.info(str(name) + " is currently not in pyspedas.")
                continue

        if delete:
            pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras']['annotations'] = None
            continue

        if pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras'].get('annotations') is None:
            pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras']['annotations'] = [annotations]
        else:
            pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras']['annotations'].append(annotations)



import math
import numpy as np
from scipy.ndimage import shift

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum


def spd_pgs_make_phi_spec(data_in, resolution=None):
    """
    Builds phi (longitudinal) spectrogram from the particle data structure

    Parameters
    ----------
        data_in: dict
            Particle data structure

        resolution: int
            Number of phi bins in the output

    Returns
    -------
    tuple
        Tuple containing: (phi values for y-axis, spectrogram values)
    """

    dr = math.pi/180.

    data = data_in.copy()

    # zero inactive bins to ensure areas with no data are represented as NaN
    zero_bins = np.argwhere(data['bins'] == 0)
    if zero_bins.size != 0:
        for item in zero_bins:
            data['data'][item[0], item[1]] = 0.0

    # get number of phi values
    if resolution is None:
        # method taken from the IDL code
        idx = np.nanargmin(np.abs((data['theta'][0, :])))
        n_phi = len(np.argwhere(data['theta'][0, :] == np.abs((data['theta'][0, :]))[idx]))
    else:
        n_phi = resolution

    # init this sample's piece of the spectrogram
    ave = np.zeros(n_phi)

    # form grid specifying the spectrogram's phi bins
    phi_grid = np.linspace(0, 360.0, n_phi+1)
    phi_grid_width = np.nanmedian(phi_grid - shift(phi_grid, 1))

    # get min/max of all data bins
    # keep phi in [0, 360]
    phi_min = (data['phi'] - 0.5*data['dphi']) 
    phi_max = (data['phi'] + 0.5*data['dphi']) % 360.0

    # algorithm below assumes maximums at 360 not wrapped to 0
    phi_max[phi_max == 0] = 360.0

    # keep phi in [0, 360]
    phi_min[phi_min < 0] = phi_min[phi_min < 0] + 360

    # keep track of bins that span phi=0
    wrapped = phi_min > phi_max

    # When averaging data bins will be weighted by the solid angle of their overlap,
    # with the given spectrogram bin.  Since each spectrogram bin spans all theta 
    # values the theta portion of that calculation can be done in advance.  These
    # values will later be multiplied by the overlap along phi to get the total 
    # solid angle. 
    omega_part = np.abs(np.sin(dr * (data['theta'] + .5*data['dtheta'])) - np.sin(dr * (data['theta'] - .5*data['dtheta'])))

    omega_part_flat = omega_part.flatten(order='F')
    phi_max_flat = phi_max.flatten(order='F')
    phi_min_flat = phi_min.flatten(order='F')
    wrapped_flat = wrapped.flatten(order='F')
    data_flat = data['data'].flatten(order='F')
    bins_flat = data['bins'].flatten(order='F')
    dphi_flat = data['dphi'].flatten(order='F')

    # Loop over each phi bin in the spectrogram and determine which data bins
    # overlap.  All overlapping bins will be weighted according to the solid 
    # angle of their intersection and averaged.
    for i in range(0, n_phi):
        weight = np.zeros(phi_min_flat.shape)

        # data bins whose maximum overlaps the current spectrogram bin
        idx_max = np.argwhere((phi_max_flat > phi_grid[i]) & (phi_max_flat < phi_grid[i+1]))
        if idx_max.size != 0:
            weight[idx_max] = (phi_max_flat[idx_max] - phi_grid[i]) * omega_part_flat[idx_max]

        # data bins whose minimum overlaps the current spectrogram bin
        idx_min = np.argwhere((phi_min_flat > phi_grid[i]) & (phi_min_flat < phi_grid[i+1]))
        if idx_min.size != 0:
            weight[idx_min] = (phi_grid[i+1] - phi_min_flat[idx_min]) * omega_part_flat[idx_min]

        # data bins contained within the current spectrogram bin
        contained = np.intersect1d(idx_max, idx_min)
        if contained.size != 0:
            weight[contained] = dphi_flat[contained] * omega_part_flat[contained]
        
        # data bins that completely cover the current spectrogram bin 
        idx_all = np.argwhere(((phi_min_flat <= phi_grid[i]) & (phi_max_flat >= phi_grid[i+1])) |
                              (wrapped_flat & ((phi_min_flat > phi_grid[i+1]) & (phi_max_flat > phi_grid[i+1]))) |
                              (wrapped_flat & ((phi_min_flat < phi_grid[i]) & (phi_max_flat < phi_grid[i]))))
        if idx_all.size != 0:
            weight[idx_all] = phi_grid_width * omega_part_flat[idx_all]

        # combine indices
        idx = np.unique(np.concatenate((idx_min, idx_max, idx_all)))

        # assign a weighted average to this bin
        if idx_max.size + idx_min.size + idx_all.size > 0:
            # normalize weighting to selected, active bins
            weight[idx] = weight[idx] * bins_flat[idx]
            weight = weight/nansum(weight)

            # average
            ave[i] = nansum(data_flat[idx]*weight[idx])

    # get y axis
    y = (phi_grid+shift(phi_grid, 1))/2.0
    y = y[1:]

    return (y, ave)



"""Print the version number for the current installation."""
import logging
# We don't want to shadow the importlib version() with pyspedas(version)
from importlib.metadata import version as imp_version
from importlib.metadata import PackageNotFoundError

def version():
    """
    Print the pyspedas version number.

    Returns
    -------
    None.

    """
    try:
        ver = imp_version("pyspedas")
    except PackageNotFoundError:
        ver = 'bleeding edge'
    logging.info("pyspedas version: " + ver)



import logging
import numpy as np
from .spd_cal_rot import spd_cal_rot


def slice2d_rotate(rotation=None, vectors=None, bfield=None, vbulk=None, sunvec=None):
    """

    """

    # Check for presence of required support data
    req_bfield = ['bv', 'be', 'perp', 'perp2', 'perp_xy', 'perp_xz', 'perp_yz', 'b_exb', 'perp1-perp2']
    req_vbulk = ['bv', 'be', 'xvel', 'perp', 'perp2', 'b_exb', 'perp1-perp2']

    if bfield is None and rotation in req_bfield:
        logging.error('Rotation: ' + rotation + ' requires B-field data')
        return

    if vbulk is None and rotation in req_vbulk:
        logging.error('Rotation: ' + rotation + ' requires bulk velocity data')
        return

    if rotation == 'bv':
        matrix = spd_cal_rot(bfield, vbulk)
    elif rotation == 'be':
        # [B, BxV] (this is the parallel - perp 2 plane)
        matrix = spd_cal_rot(bfield, np.cross(bfield, vbulk))
    elif rotation == 'xy':
        matrix = spd_cal_rot([1, 0, 0], [0, 1, 0])
    elif rotation == 'xz':
        matrix = spd_cal_rot([1, 0, 0], [0, 0, 1])
    elif rotation == 'yz':
        matrix = spd_cal_rot([0, 1, 0], [0, 0, 1])
    elif rotation == 'xvel':
        matrix = spd_cal_rot([1, 0, 0], vbulk)
    elif rotation == 'perp':
        # perp 2 - perp 1 plane; should be: [(BxV)xB, BxV]
        matrix = spd_cal_rot(np.cross(np.cross(bfield, vbulk), bfield), np.cross(bfield, vbulk))
    elif rotation == 'perp1-perp2':
        # perp 1 - perp 2 plane; should be: [BxV, (BxV)xB]
        matrix = spd_cal_rot(np.cross(bfield, vbulk), np.cross(np.cross(bfield, vbulk), bfield))
    elif rotation == 'perp_yz':
        matrix = spd_cal_rot(np.cross(np.cross(bfield, [0, 1, 0]), bfield), np.cross(np.cross(bfield, [0, 0, 1]), bfield))
    elif rotation == 'perp_xy':
        matrix = spd_cal_rot(np.cross(np.cross(bfield, [1, 0, 0]), bfield), np.cross(np.cross(bfield, [0, 1, 0]), bfield))
    elif rotation == 'perp_xz':
        matrix = spd_cal_rot(np.cross(np.cross(bfield, [1, 0, 0]), bfield), np.cross(np.cross(bfield, [0, 0, 1]), bfield))
    elif rotation == 'b_exb':
        # [B, (BxV)xB] (this is the parallel - perp 2 plane)
        matrix = spd_cal_rot(bfield, np.cross(np.cross(bfield, vbulk), bfield))
    else:
        logging.error('Unknown rotation: ' + rotation)
        return

    if rotation != 'xy':
        logging.info('Aligning slice plane to: ' + rotation)

    # Transform particle and support vectors
    if vectors is not None:
        transformed = np.zeros(vectors.shape)
        num_vectors = vectors.shape[0]
        for vector_idx in range(num_vectors):
            transformed[vector_idx] = (matrix.T @ vectors[vector_idx, :]).flatten()
        vectors = transformed
    if vbulk is not None:
        vbulk = matrix.T @ vbulk
    if bfield is not None:
        bfield = matrix.T @ bfield
    if sunvec is not None:
        sunvec = matrix.T @ sunvec

    return {'matrix': matrix, 'vectors': vectors, 'bfield': bfield, 'vbulk': vbulk, 'sunvec': sunvec}



import numpy as np
from pyspedas.tplot_tools import get, store
from pyspedas.tplot_tools import tnames


def mms_feeps_flat_field_corrections(probes = ['1', '2', '3', '4'], data_rate = 'brst', suffix = ''):
    """
       Apply flat field correction factors to FEEPS ion/electron data;
       correct factors are from the gain factor found in:
       
           FlatFieldResults_V3.xlsx
           
       from Drew Turner, 1/19/2017
    
    Parameters
    ---------------
        probes: list of str
            list of probes #, e.g., '4' for MMS4

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        suffix: str
            suffix of the loaded data

    Notes
    ---------------
        From Drew Turner, 1/18/17:
           Here are the correction factors that we need to apply to the current 
           ION counts/rates/fluxes in the CDF files.  
           NOTE, THIS IS A DIFFERENT TYPE OF CORRECTION THAN THAT FOR THE ELECTRONS!  
           These shifts should be applied to the counts/rates/fluxes data EYE-BY-EYE on each spacecraft.  
           These are multiplication factors (i.e., Jnew = Jold * Gcorr). 
           For those equations, Jold is the original count/rate/flux array and
           Jnew is the corrected version of the arrays using the factors listed below.
           
            MMS1:
            Top6: Gcorr = 0.7
            Top7: Gcorr = 2.5
            Top8: Gcorr = 1.5
            Bot6: Gcorr = 0.9
            Bot7: Gcorr = 1.2
            Bot8: Gcorr = 1.0

            MMS2:
            Top6: Gcorr = 1.3
            Top7: BAD EYE
            Top8: Gcorr = 0.8
            Bot6: Gcorr = 1.4
            Bot7: BAD EYE
            Bot8: Gcorr = 1.5

            MMS3:
            Top6: Gcorr = 0.7
            Top7: Gcorr = 0.8
            Top8: Gcorr = 1.0
            Bot6: Gcorr = 0.9
            Bot7: Gcorr = 0.9
            Bot8: Gcorr = 1.3

            MMS4:
            Top6: Gcorr = 0.8
            Top7: BAD EYE
            Top8: Gcorr = 1.0
            Bot6: Gcorr = 0.8
            Bot7: Gcorr = 0.6
            Bot8: Gcorr = 0.9
    """
    
    G_corr = {}
    G_corr['mms1-top6'] = 0.7
    G_corr['mms1-top7'] = 2.5
    G_corr['mms1-top8'] = 1.5
    G_corr['mms1-bot5'] = 1.2 # updated 1/24
    G_corr['mms1-bot6'] = 0.9
    G_corr['mms1-bot7'] = 2.2 # updated 1/24
    G_corr['mms1-bot8'] = 1.0

    G_corr['mms2-top4'] = 1.2 # added 1/24
    G_corr['mms2-top6'] = 1.3
    G_corr['mms2-top7'] = 0 # bad eye
    G_corr['mms2-top8'] = 0.8
    G_corr['mms2-bot6'] = 1.4
    G_corr['mms2-bot7'] = 0 # bad eye
    G_corr['mms2-bot8'] = 1.5

    G_corr['mms3-top6'] = 0.7
    G_corr['mms3-top7'] = 0.8
    G_corr['mms3-top8'] = 1.0
    G_corr['mms3-bot6'] = 0.9
    G_corr['mms3-bot7'] = 0.9
    G_corr['mms3-bot8'] = 1.3

    G_corr['mms4-top6'] = 0.8
    G_corr['mms4-top7'] = 0 # bad eye
    G_corr['mms4-top8'] = 1.0
    G_corr['mms4-bot6'] = 0.8
    G_corr['mms4-bot7'] = 0.6
    G_corr['mms4-bot8'] = 0.9
    G_corr['mms4-bot9'] = 1.5 # added 1/24

    #sensor_ids = ['6', '7', '8']
    sensor_ids = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12']
    sensor_types = ['top', 'bottom']
    levels = ['l2']
    species = ['ion', 'electron']

    for probe in probes:
        for sensor_type in sensor_types:
            for sensor_id in sensor_ids:
                if G_corr.get('mms'+probe+'-'+sensor_type[0:3]+sensor_id) is not None:
                    correction = G_corr['mms'+probe+'-'+sensor_type[0:3]+sensor_id]
                else:
                    correction = 1.0

                for level in levels:
                    for species_id in species:
                        if correction != 1.0:
                            cr_var = tnames('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+species_id+'_'+sensor_type+'_count_rate_sensorid_'+sensor_id+suffix)
                            i_var = tnames('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+species_id+'_'+sensor_type+'_intensity_sensorid_'+sensor_id+suffix)
                            c_var = tnames('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+species_id+'_'+sensor_type+'_counts_sensorid_'+sensor_id+suffix)

                            if not cr_var:
                                count_rate = None
                            else:
                                count_rate = get(cr_var[0])
                            if count_rate is not None:
                                cr_times, cr_data, cr_energies = count_rate
                                if np.isnan(cr_energies).all():
                                    continue
                                store(cr_var[0], data={'x': cr_times, 'y': cr_data*correction, 'v': cr_energies})

                            if not i_var:
                                intensity = None
                            else:
                                intensity = get(i_var[0])
                            if intensity is not None:
                                i_times, i_data, i_energies = intensity
                                if np.isnan(i_energies).all():
                                    continue
                                store(i_var[0], data={'x': i_times, 'y': i_data*correction, 'v': i_energies})

                            if not c_var:
                                counts = None
                            else:
                                counts = get(c_var[0])
                            if counts is not None:
                                c_times, c_data, c_energies = counts
                                if np.isnan(c_energies).all():
                                    continue
                                store(c_var[0], data={'x': c_times, 'y': c_data*correction, 'v': c_energies})



import urllib

def find_ip_address():
    """ Uses api.ipify.org to get the public IP address, useful for troubleshooting data access issues in test suites"""
    url='http://api.ipify.org'
    page=urllib.request.urlopen(url)
    address=page.read().decode('utf-8')
    return address


from .load import load


def mageis(trange=['2015-11-5', '2015-11-6'],
           probe='a',
           datatype='',
           level='l3',
           rel='rel04',
           prefix='',
           suffix='',
           force_download=False,
           get_support_data=False,
           varformat=None,
           varnames=[],
           downloadonly=False,
           notplot=False,
           no_update=False,
           time_clip=False):
    """
    This function loads data from the Energetic Particle, Composition, and Thermal Plasma Suite (ECT)

    Parameters
    ----------
        trange : list of str, default=['2015-11-5', '2015-11-6']
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str, default='a'
            Spacecraft probe name: 'a' or 'b'

        datatype : str, default=''
            Data type. Valid options are specific to different data levels.

        level : str, default='l3'
            Data level. Valid options: 'l1', 'l2', 'l3', 'l4'

        rel : str, default='rel04'
            Release version of the data.

        prefix : str, optional
            The tplot variable names will be given this prefix. By default, no prefix is added.

        suffix: str, optional
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        force_download : bool, default=False
            Download file even if local version is more recent than server version.

        get_support_data: bool, default=False
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str, optional
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str, optional
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool, default=False
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool, default=False
            Return the data in hash tables instead of creating tplot variables

        no_update: bool, default=False
            If set, only load data from your local cache

        time_clip: bool, default=False
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    -------
    tvars : dict or list
        List of created tplot variables or dict of data tables if notplot is True.

    Examples
    --------
    >>> mageis_vars = pyspedas.projects.rbsp.mageis(trange=['2018-11-5', '2018-11-6'], level='l3', rel='rel04')
    >>> tplot('I')
    """
    return load(instrument='mageis', rel=rel, trange=trange, probe=probe, datatype=datatype, level=level, prefix=prefix, suffix=suffix, force_download=force_download, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)



from cdasws import CdasWs


def find_datasets(mission=None, instrument=None, label=False, quiet=False):
    """
    Find datasets from the Coordinated Data Analysis System (CDAS) based on mission and/or instrument criteria.

    This function queries the CDAS for datasets, optionally filtering by mission and instrument. It can
    also display labels for the datasets.

    Parameters
    ----------
    mission : str, optional
        The name of the mission to filter the datasets. If None, datasets are not filtered by mission.
    instrument : str, optional
        The name of the instrument to filter the datasets. If None, datasets are not filtered by instrument.
    label : bool, default=False
        If True, the function prints both the dataset ID and label. If False, only the dataset ID is printed.
    quiet: bool, default=False
        If True, suppresses printing of dataset IDs (and labels) found

    Returns
    -------
    list of str
        List of datasets found

    Examples
    --------
    >>> from pyspedas import find_datasets
    >>> find_datasets(mission='MMS', instrument='FGM')
    MMS1_FGM_BRST_L2
    MMS1_FGM_SRVY_L2
    ...

    >>> # Suppress printed output
    >>> from pyspedas import find_datasets
    >>> mms_list = find_datasets(mission='MMS', quiet=True)
    >>> print(mms_list[0:3])

    """

    cdas = CdasWs()
    datasets = cdas.get_datasets(observatoryGroup=mission)
    output_list = []
    for index, dataset in enumerate(datasets):
        if instrument is not None:
            if instrument.upper() not in dataset['Id']:
                continue
        if label:
            if not quiet:
                print(dataset['Id'] + ': ' + dataset['Label'])
        else:
            if not quiet:
                print(dataset['Id'])
        output_list.append(dataset['Id'])
    return output_list


from .load import load
from .mag_postprocessing import mag_postprocessing

# This routine was originally in st5/__init__.py, until being moved to its own file.
# Please refer to __init__.py if you need to see the revision history before it was moved.

def mag(trange=['2006-06-01', '2006-06-02'],
        probe='094',
        datatype='1sec',
        suffix='',
        prefix='',
        get_support_data=False,
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Magnetometer (MAG)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2006-06-01', '2006-06-02']

        probe: str
            Probe #; Valid options: '094', '224', '155'
            Default: '094'

        datatype: str
            Data type; options: '1sec'
            Default: '1sec'

        suffix: str
            The tplot variable names will be given this suffix.
            Default: '', no suffix is added.

        prefix: str
            The tplot variable names will be given this prefix.
            Default: '', no prefix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False. Only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None. All variables are loaded in.

        varnames: list of str
            List of variable names to load.
            Default: []. If not specified, all data variables are loaded.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

    Returns
    ----------
        List of tplot variables created.

    Example:
    ----------
        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> mag_vars = pyspedas.projects.st5.mag(trange=['2006-06-01', '2006-06-02'])
        >>> tplot(['B_SM', 'SC_POS_SM'])

    """

    tvars = load(instrument='mag', trange=trange, datatype=datatype, probe=probe, suffix=suffix, prefix=prefix,
                 get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly,
                 notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return mag_postprocessing(tvars)



"""Test functions in the utilites folder."""

import unittest
import numpy as np
from numpy.testing import assert_allclose

from pyspedas.projects import themis, mms
from pyspedas import (
    dailynames,
    tcopy,
    cotrans,
    mpause_2,
    mpause_t96,
    find_datasets,
    data_exists,
    tkm2re,
    split_vec,
    del_data,
    time_clip,
    get_data,
    store_data,
    options,
)


class UtilTestCases(unittest.TestCase):
    """Test fuctions in the utilites folder."""

    def test_dailynames(self):
        """Test dailynames function."""
        self.assertTrue(dailynames(trange=["2015-12-1", "2015-12-1/2:00"], hour_res=True) == ["2015120100", "2015120101"])
        self.assertTrue(dailynames(trange=["2015-12-1", "2015-12-3"]) == ["20151201", "20151202"])
        self.assertTrue(dailynames(trange=["2015-12-3", "2015-12-2"]) == ["20151203"])
        self.assertTrue(dailynames() is None)
        self.assertTrue(
            dailynames(trange=["2015-12-3", "2019-12-2"], file_format="%Y") == ["2015", "2016", "2017", "2018", "2019"]
        )
        self.assertTrue(dailynames(trange=["2015-1-1", "2015-3-2"], file_format="%Y%m") == ["201501", "201502", "201503"])
        self.assertTrue(
            dailynames(trange=["2015-1-1", "2015-3-2"], file_format="/%Y/%m/") == ["/2015/01/", "/2015/02/", "/2015/03/"]
        )
        self.assertTrue(dailynames(trange=["2015-1-1", "2015-1-1/3:00"], file_format="%H", res=60.0) == ["00", "01", "02"])
        self.assertTrue(
            dailynames(trange=["2015-1-1/2:00", "2015-1-1/3:00"], file_format="%M", res=600.0)
            == ["00", "10", "20", "30", "40", "50"]
        )

    def test_tcopy(self):
        """Test tcopy function."""
        store_data("test", data={"x": [1, 2, 3], "y": [5, 5, 5]})
        tcopy("test")
        tcopy("test", "another-copy")
        t, d = get_data("test-copy")
        self.assertTrue(t.tolist() == [1, 2, 3])
        self.assertTrue(d.tolist() == [5, 5, 5])
        t, d = get_data("another-copy")
        self.assertTrue(t.tolist() == [1, 2, 3])
        self.assertTrue(d.tolist() == [5, 5, 5])
        # the following should gracefully error
        tcopy("doesnt exist", "another-copy")
        tcopy(["another-copy", "test"], "another-copy")

    def test_tkm2re(self):
        store_data("test", data={"x": [1, 2, 3], "y": [5, 5, 5]})
        options("test", "ysubtitle", "[Re]")
        # convert to km
        tkm2re("test", km=True)
        # convert back
        tkm2re("test_km")
        self.assertTrue(data_exists("test_km_re"))
        nothing = tkm2re("doesnt_exist")
        self.assertTrue(nothing is None)
        tkm2re("test_km", newname="another_test_km")
        self.assertTrue(data_exists("another_test_km"))
        anerror = tkm2re("test_km", newname=["test1_km", "test1_km"])
        self.assertTrue(anerror is None)

    def test_time_clip(self):
        import pyspedas

        x = [1, 2, 3, 4, 5]
        y = [2, 4, 6, 8, 10]
        xfp = [1.0, 2.0, 3.0, 4.0, 5.0]
        # Data for a scalar variable
        y_scalar = [0.0, 1.0, 2.0, 3.0, 4.0]
        # Data for a vector variable
        y_vec = [
            [0.0, 1.0, 2.0],
            [3.0, 4.0, 5.0],
            [6.0, 7.0, 8.0],
            [9.0, 10.0, 11.0],
            [12.0, 13.0, 14.0],
        ]
        y_vec = np.array(y_vec)
        # Constant metadata for a vector variable
        v1_1d = [1.0, 2.0, 3.0]
        # Time varying metadata for a vector variable
        v1_2d = y_vec
        # Data for a 3x3 matrix variable
        y_mat = [
            [[0.0, 1.0, 2.0], [0.0, 1.0, 2.0], [0.0, 1.0, 2.0]],
            [[3.0, 4.0, 5.0], [3.0, 4.0, 5.0], [3.0, 4.0, 5.0]],
            [[6.0, 7.0, 8.0], [6.0, 7.0, 8.0], [6.0, 7.0, 8.0]],
            [[9.0, 10.0, 11.0], [9.0, 10.0, 11.0], [9.0, 10.0, 11.0]],
            [[12.0, 13.0, 14.0], [12.0, 13.0, 14.0], [12.0, 13.0, 14.0]],
        ]
        y_mat = np.array(y_mat)
        v2_1d = v1_1d
        v2_2d = y_vec

        # Scalar
        store_data("fptest", data={"x": xfp, "y": y_scalar})
        # 1-D vec, no metadata
        store_data("fptest_vec_no_v1", data={"x": xfp, "y": y_vec})
        # 1-D vec, with 1-D (constant) V1
        store_data("fptest_vec_v1_1d", data={"x": xfp, "y": y_vec, "v": v1_1d})
        # 1-D vec, with 2-D (time varying) V1
        store_data("fptest_vec_v1_2d", data={"x": xfp, "y": y_vec, "v": v1_2d})
        # 3x3 mat, with no V1, no V2
        store_data("fptest_mat_v1_none_v2_none", data={"x": xfp, "y": y_mat})
        # 3x3 mat, with 1-D (constant) V1, 1-D (constant) V2
        store_data(
            "fptest_mat_v1_1d_v2_1d",
            data={"x": xfp, "y": y_mat, "v1": v1_1d, "v2": v2_1d},
        )
        # 3x3 mat, with 2-D (time varying) V1, 2-D (time varying) V2
        store_data(
            "fptest_mat_v1_2d_v2_2d",
            data={"x": xfp, "y": y_mat, "v1": v1_2d, "v2": v2_2d},
        )
        # 3x3 mat, with 2-D (time varying) V1, 1-D (constant) V2
        store_data(
            "fptest_mat_v1_2d_v2_1d",
            data={"x": xfp, "y": y_mat, "v1": v1_2d, "v2": v2_1d},
        )
        # 3x3 mat, with 1-D (constant) V1, 2-D (time varying) V2
        store_data(
            "fptest_mat_v1_1d_v2_2d",
            data={"x": xfp, "y": y_mat, "v1": v1_1d, "v2": v2_2d},
        )
        # 3x3 mat, with missing V1, 2-D (time varying) V2
        # This doesn't quite work.  The 'v2' value doesn't survive the store_data/get_data round trip
        # if v1 is missing.  Not sure if that really qualifies as a bug, considering how ill-advised such a
        # construct would be...
        # store_data('fptest_mat_v1_none_v2_2d',data={'x':xfp,'y':y_mat, 'v1':None, 'v2':v2_2d})

        # Test warning for no data in time range
        time_clip("fptest", 1.5, 1.7)

        # Single value in time range
        time_clip("fptest", 1.5, 2.5)
        self.assertTrue(data_exists("fptest-tclip"))
        dat = get_data("fptest-tclip")
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)

        # Multiple values in time range
        time_clip("fptest", 1.5, 3.5)
        dat = get_data("fptest-tclip")
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, [2.0, 3.0])

        # Multiple values in time range -- clip interior
        time_clip("fptest", 1.5, 3.5, suffix="-tclip-interior", interior_clip=True)
        dat = get_data("fptest-tclip-interior")
        self.assertTrue(len(dat.times) == 3)
        assert_allclose(dat.times, [1.0, 4.0, 5.0])

        # Entire time range
        time_clip("fptest", -100.0, 100.0, suffix="-full_range")
        self.assertTrue(data_exists("fptest-full_range"))
        dat = get_data("fptest-full_range")
        self.assertTrue(len(dat.times) == 5)
        assert_allclose(dat.times, xfp)

        # Vector data, single value in time range, no metadata
        time_clip("fptest_vec_no_v1", 1.5, 2.5)
        self.assertTrue(data_exists("fptest_vec_no_v1-tclip"))
        dat = get_data("fptest_vec_no_v1-tclip")
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        assert_allclose(dat.y[0, :], [3.0, 4.0, 5.0])

        # Vector data, single value in time range, constant metadata
        time_clip("fptest_vec_v1_1d", 1.5, 2.5)
        self.assertTrue(data_exists("fptest_vec_v1_1d-tclip"))
        dat = get_data("fptest_vec_v1_1d-tclip")
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        assert_allclose(dat.y[0, :], [3.0, 4.0, 5.0])
        assert_allclose(dat.v, v1_1d)

        # Vector data, multiple values in time range, constant metadata
        time_clip("fptest_vec_v1_1d", 1.5, 3.5)
        self.assertTrue(data_exists("fptest_vec_v1_1d-tclip"))
        dat = get_data("fptest_vec_v1_1d-tclip")
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, [2.0, 3.0])
        assert_allclose(dat.y, [[3.0, 4.0, 5.0], [6.0, 7.0, 8.0]])
        assert_allclose(dat.v, v1_1d)

        # Vector data, single value in time range, time-varying metadata
        time_clip("fptest_vec_v1_2d", 1.5, 2.5)
        self.assertTrue(data_exists("fptest_vec_v1_2d-tclip"))
        dat = get_data("fptest_vec_v1_2d-tclip")
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        self.assertTrue(dat.v.shape == (1, 3))
        assert_allclose(dat.y[0, :], [3.0, 4.0, 5.0])
        assert_allclose(dat.v, v1_2d[1:2])

        # Vector data, multiple values in time range, time-varying metadata
        time_clip("fptest_vec_v1_2d", 1.5, 3.5)
        self.assertTrue(data_exists("fptest_vec_v1_2d-tclip"))
        dat = get_data("fptest_vec_v1_2d-tclip")
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, xfp[1:3])
        self.assertTrue(dat.y.shape == (2, 3))
        assert_allclose(dat.y, y_vec[1:3, :])
        self.assertTrue(dat.v.shape == (2, 3))
        assert_allclose(dat.v, v1_2d[1:3])

        # matrix data, single value in time range, no metadata
        time_clip("fptest_mat_v1_none_v2_none", 1.5, 2.5)
        self.assertTrue(data_exists("fptest_mat_v1_none_v2_none-tclip"))
        dat = get_data("fptest_mat_v1_none_v2_none-tclip")
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        assert_allclose(dat.y[0, :, :], y_mat[1, :, :])

        # matrix data, single value in time range, constant v1 and v2 metadata
        time_clip("fptest_mat_v1_1d_v2_1d", 1.5, 2.5)
        self.assertTrue(data_exists("fptest_mat_v1_1d_v2_1d-tclip"))
        dat = get_data("fptest_mat_v1_1d_v2_1d-tclip")
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        assert_allclose(dat.y[0, :, :], y_mat[1, :, :])
        assert_allclose(dat.v1, v1_1d)
        assert_allclose(dat.v2, v2_1d)

        # matrix data, multiple values in time range, constant v1 and v2 metadata
        time_clip("fptest_mat_v1_1d_v2_1d", 1.5, 3.5)
        self.assertTrue(data_exists("fptest_mat_v1_1d_v2_1d-tclip"))
        dat = get_data("fptest_mat_v1_1d_v2_1d-tclip")
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, [2.0, 3.0])
        assert_allclose(dat.y, y_mat[1:3, :, :])
        assert_allclose(dat.v1, v1_1d)
        assert_allclose(dat.v2, v2_1d)

        # matrix data, single value in time range, constant v1, time-varying v2 metadata
        time_clip("fptest_mat_v1_1d_v2_2d", 1.5, 2.5)
        self.assertTrue(data_exists("fptest_mat_v1_1d_v2_2d-tclip"))
        dat = get_data("fptest_mat_v1_1d_v2_2d-tclip")
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        self.assertTrue(dat.v1.shape == (3,))
        assert_allclose(dat.y[:, :, :], y_mat[1:2, :, :])
        assert_allclose(dat.v1, v1_1d)
        assert_allclose(dat.v2, v1_2d[1:2])

        # matrix data, multiple values in time range, constant v1, time-varying v2 metadata
        time_clip("fptest_mat_v1_1d_v2_2d", 1.5, 3.5)
        self.assertTrue(data_exists("fptest_mat_v1_1d_v2_2d-tclip"))
        dat = get_data("fptest_mat_v1_1d_v2_2d-tclip")
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, xfp[1:3])
        self.assertTrue(dat.v1.shape == (3,))
        assert_allclose(dat.y[:, :, :], y_mat[1:3, :, :])
        assert_allclose(dat.v1, v1_1d)
        assert_allclose(dat.v2, v1_2d[1:3])

        # matrix data, single values in time range, time-varying v1, constant v2 metadata
        time_clip("fptest_mat_v1_2d_v2_1d", 1.5, 2.5)
        self.assertTrue(data_exists("fptest_mat_v1_2d_v2_1d-tclip"))
        dat = get_data("fptest_mat_v1_2d_v2_1d-tclip")
        self.assertTrue(len(dat.times) == 1)
        self.assertTrue(dat.times[0] == 2.0)
        self.assertTrue(dat.v1.shape == (1, 3))
        self.assertTrue(dat.v2.shape == (3,))
        assert_allclose(
            dat.y[
                :,
                :,
            ],
            y_mat[1:2, :, :],
        )
        assert_allclose(dat.v1, v1_2d[1:2])
        assert_allclose(dat.v2, v2_1d)

        # matrix data, multiple values in time range, time-varying v1, constant v2 metadata
        time_clip("fptest_mat_v1_2d_v2_1d", 1.5, 3.5)
        self.assertTrue(data_exists("fptest_mat_v1_2d_v2_1d-tclip"))
        dat = get_data("fptest_mat_v1_2d_v2_1d-tclip")
        self.assertTrue(len(dat.times) == 2)
        assert_allclose(dat.times, [2.0, 3.0])
        self.assertTrue(dat.v1.shape == (2, 3))
        self.assertTrue(dat.v2.shape == (3,))
        assert_allclose(
            dat.y[
                :,
                :,
            ],
            y_mat[1:3, :, :],
        )
        assert_allclose(dat.v1, v1_2d[1:3])
        assert_allclose(dat.v2, v2_1d)

        # Multiple variables in one call
        store_data("tst1", data={"x": x, "y": y})
        store_data("tst2", data={"x": x, "y": y})
        # reversed time limits
        time_clip(["tst1", "tst2"], 10, 1)
        # data completely outside time limits
        time_clip(["tst1", "tst2"], 10, 20)
        # one point in limits
        time_clip("tst1", 1.5, 2.5)
        self.assertTrue(data_exists("tst1-tclip"))
        # overwrite
        del_data("tst1-tclip")
        time_clip("tst1", 1.5, 2.0, overwrite=True)
        self.assertFalse(data_exists("tst1-tclip"))
        time_clip("tst1", 1.5, 2.5, newname="tst1_new")
        self.assertTrue(data_exists("tst1_new"))
        time_clip("tst1", 1.5, 2.5, newname="", suffix="-tc1")
        self.assertTrue(data_exists("tst1-tc1"))
        time_clip("tst1", 1.5, 2.5, newname=[], suffix="-tc2")
        self.assertTrue(data_exists("tst1-tc2"))
        # newname has different count than input names, default to suffix
        time_clip("tst1", 1.5, 2.5, newname=["foo", "bar"], suffix="-tc3")
        self.assertTrue(data_exists("tst1-tc3"))
        time_clip("tst1", 1.5, 2.5, newname=None, suffix="-tc4")
        self.assertTrue(data_exists("tst1-tc4"))
        # no such tplot name
        time_clip("bogus", 1.5, 2.5)
        # empty input list
        time_clip([], 1.5, 2.5)

        # Strings as start/end times
        trange = ["2007-03-23/9:30", "2007-03-23/12:00"]
        # Note times are out of order if compared as strings!
        self.assertTrue(trange[0] > trange[1])
        themis.state(probe="a", trange=trange)
        time_clip("tha_pos", trange[0], trange[1], newname="tha_pos-tclip")
        self.assertTrue(data_exists("tha_pos-tclip"))
        dat = get_data("tha_pos-tclip")
        self.assertTrue(dat.times[0] >= pyspedas.time_float(trange[0]))

    def test_mpause_t96(self):
        trange = ["2023-03-24", "2023-03-25"]
        # THEMIS orbits come from the 'state' datatype
        # We will use GSE coordinates for the plots we'll make.
        themis.state(probe="a", trange=trange, varformat="*_pos_gse")
        cotrans("tha_pos_gse", name_out="tha_pos_gsm", coord_in="GSE", coord_out="GSM")
        posdat = get_data("tha_pos_gsm")
        re = 6378.0
        pos_gsm = posdat.y / re
        xmgnp, ymgnp, zmgnp, id, distan = mpause_t96(pd=14.0, xgsm=pos_gsm[:, 0], ygsm=pos_gsm[:, 1], zgsm=pos_gsm[:, 2])
        self.assertEqual(len(xmgnp), 90)
        self.assertEqual(id.min(), -1)
        self.assertEqual(id.max(), 1)
        self.assertTrue(distan.min() < 0.01)
        self.assertTrue(distan.max() > 10.0)
        self.assertEqual(len(id), len(posdat.times))

    def test_mpause_2(self):
        trange = ["2023-03-24", "2023-03-25"]
        # THEMIS orbits come from the 'state' datatype
        # We will use GSE coordinates for the plots we'll make.
        themis.state(probe="a", trange=trange, varformat="*_pos_gse")
        cotrans("tha_pos_gse", name_out="tha_pos_gsm", coord_in="GSE", coord_out="GSM")
        posdat = get_data("tha_pos_gsm")
        re = 6378.0
        pos_gsm = posdat.y / re
        self.assertIsNotNone(pos_gsm)
        xmp, ymp = mpause_2()
        self.assertEqual(len(xmp), 2000)
        self.assertEqual(len(ymp), 2000)
        self.assertTrue(xmp.min() < -299.0)
        self.assertTrue(xmp.max() > 10.77)
        self.assertTrue(ymp.min() < -68.0)
        self.assertTrue(ymp.max() > 77.8)

    def test_find_datasets_quiet(self):
        ds = find_datasets(mission="MMS", instrument="FGM", quiet=True)
        self.assertTrue("MMS1_FGM_BRST_L2" in ds)
        self.assertTrue("MMS2_FGM_BRST_L2" in ds)

    def test_find_datasets_nolabel(self):
        ds = find_datasets(mission="MMS", instrument="FGM")
        self.assertTrue("MMS1_FGM_BRST_L2" in ds)
        self.assertTrue("MMS2_FGM_BRST_L2" in ds)

    def test_find_datasets_label(self):
        ds = find_datasets(mission="MMS", instrument="FGM", label=True)
        self.assertTrue("MMS1_FGM_BRST_L2" in ds)
        self.assertTrue("MMS2_FGM_BRST_L2" in ds)

    def test_split_vec_metadata(self):
        del_data("*")
        themis.fit(probe="c")
        md = get_data("thc_fgs_dsl", metadata=True)
        split_vec("thc_fgs_dsl")
        self.assertTrue(data_exists("thc_fgs_dsl_x"))
        self.assertTrue(data_exists("thc_fgs_dsl_y"))
        self.assertTrue(data_exists("thc_fgs_dsl_z"))
        md_x = get_data("thc_fgs_dsl_x", metadata=True)
        self.assertTrue(md_x["plot_options"]["xaxis_opt"]["axis_label"] == md["plot_options"]["xaxis_opt"]["axis_label"])
        self.assertTrue(md_x["plot_options"]["yaxis_opt"]["legend_names"] == md["CDF"]["LABELS"][0:1])
        md_y = get_data("thc_fgs_dsl_y", metadata=True)
        self.assertTrue(md_y["plot_options"]["xaxis_opt"]["axis_label"] == md["plot_options"]["xaxis_opt"]["axis_label"])
        self.assertTrue(md_y["plot_options"]["yaxis_opt"]["legend_names"] == md["CDF"]["LABELS"][1:2])

    def test_imports(self):
        # fully qualified name without .projects.
        del_data("*")
        themis.state(probe="a")
        self.assertTrue(data_exists("tha_pos"))
        # fully qualified name with .projects.
        del_data("*")
        themis.state(probe="a")
        themis.state(probe="a")
        self.assertTrue(data_exists("tha_pos"))

        del_data("*")
        themis.state(probe="a")
        self.assertTrue(data_exists("tha_pos"))

        del_data("*")
        themis.state(probe="a")
        self.assertTrue(data_exists("tha_pos"))

        del_data("*")
        mms.fgm()
        self.assertTrue(data_exists("mms1_fgm_b_gsm_srvy_l2"))

        del_data("*")
        themis.state(probe="a")
        self.assertTrue(data_exists("tha_pos"))


if __name__ == "__main__":
    unittest.main()



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG

def load(trange=['1997-01-03', '1997-01-04'], 
         instrument='mfe',
         datatype='k0',
         suffix='', 
         prefix='',
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         force_download=False):
    """
    This function loads data from the Polar mission; this function is not meant 
    to be called directly; instead, see the wrappera:
        pyspedas.projects.polar.mfe
        pyspedas.projects.polar.efi
        pyspedas.projects.polar.pwi
        pyspedas.projects.polar.hydra
        pyspedas.projects.polar.tide
        pyspedas.projects.polar.timas
        pyspedas.projects.polar.cammice
        pyspedas.projects.polar.ceppad
        pyspedas.projects.polar.uvi
        pyspedas.projects.polar.vis
        pyspedas.projects.polar.pixie
        pyspedas.projects.polar.orbit

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['1997-01-03', '1997-01-04']

        instrument: str
            Spacecraft identifier ('mfe', 'efi', 'pwi', 'hydra', 'tide', 'timas',
                'cammice', 'ceppad', 'uvi', 'vis', 'pixie', 'orbit')
            Default: 'mfe'

        datatype: str
            Valid options: 'k0'
            Default: 'k0'

        suffix: str
            The tplot variable names will be given this suffix. By default,
            no prefix is added.
            Default: ''
        
        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: 'False', only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None, all variables are loaded in.

        varnames: list of str
            List of variable names to load
            Default: [], all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            If True, downloads the file even if a newer version exists locally. 
            Default: False.

    Returns
    ----------
        List of tplot variables created.

    Example
    ----------
        import pyspedas
        from pyspedas.tplot_tools import tplot
        mfe_polar_vars = pyspedas.projects.polar.mfe(trange=['1997-01-03', '1997-01-04'])

        efi_polar_vars = pyspedas.projects.polar.efi(trange=['1997-01-03', '1997-01-04'])

        pwi_polar_vars = pyspedas.projects.polar.pwi(trange=['1997-01-03', '1997-01-04'])

        tide_polar_vars = pyspedas.projects.polar.tide(trange=['1997-01-03', '1997-01-04'])

        orbit_polar_vars = pyspedas.projects.polar.orbit(trange=['1997-01-03', '1997-01-04'])

    """
    if prefix is None:
        prefix = ''
    
    if suffix is None:
        suffix = ''

    if instrument == 'mfe':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'efi':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'pwi':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'hydra':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_hyd_%Y%m%d_v??.cdf'
    elif instrument == 'tide':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_tid_%Y%m%d_v??.cdf'
    elif instrument == 'timas':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_tim_%Y%m%d_v??.cdf'
    elif instrument == 'cammice':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_cam_%Y%m%d_v??.cdf'
    elif instrument == 'ceppad':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_cep_%Y%m%d_v??.cdf'
    elif instrument == 'uvi':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'vis':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
    elif instrument == 'pixie':
        pathformat = instrument+'/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_pix_%Y%m%d_v??.cdf'
    elif instrument == 'spha':
        pathformat = 'orbit/'+instrument+'_'+datatype+'/%Y/po_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, prefix=prefix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import unittest
from pyspedas import kompsat_load
from pyspedas.tplot_tools import data_exists


class Kompsat_Tests(unittest.TestCase):
    def test_mag(self):
        var_names = kompsat_load(trange=["2021-01-01 02:00", "2021-01-01 03:00"])
        self.assertTrue(data_exists("sosmag_b_gse"))
        self.assertTrue("sosmag_position" in var_names)

    def test_1m(self):
        var_names = kompsat_load(
            datatype="1m", trange=["2021-01-01 02:00", "2021-01-01 03:00"]
        )
        self.assertTrue(data_exists("sosmag_1m_b_gse"))
        self.assertTrue(data_exists("sosmag_1m_position"))
        self.assertTrue("sosmag_1m_b_gse" in var_names)

    def test_1m_get_support_data(self):
        var_names = kompsat_load(
            datatype="1m", trange=["2021-01-01 02:00", "2021-01-01 03:00"], get_support_data=True
        )
        self.assertTrue(data_exists("sosmag_1m_b_gse"))
        self.assertTrue(data_exists("sosmag_1m_position"))
        self.assertTrue("sosmag_1m_b_gse" in var_names)
        self.assertTrue(data_exists("sosmag_1m_data_flags"))
        self.assertTrue("sosmag_1m_data_flags" in var_names)

    def test_p(self):
        var_names = kompsat_load(
            instrument="p", trange=["2024-04-01 02:00", "2024-04-01 03:00"]
        )
        self.assertTrue(data_exists("kompsat_p_p1"))
        self.assertTrue(data_exists("kompsat_p_p8"))
        self.assertTrue("kompsat_p_p5" in var_names)

    def test_e(self):
        var_names = kompsat_load(
            instrument="e",
            trange=["2024-04-01 02:00", "2024-04-01 03:00"],
            get_support_data=True,
        )
        self.assertTrue(data_exists("kompsat_e_e1"))
        self.assertTrue(data_exists("kompsat_e_e10_qef"))
        self.assertTrue("kompsat_e_e6" in var_names)


if __name__ == "__main__":
    unittest.main()



"""
Creates a new tplot variable as the time average of original.

    Notes
    -----
    Similar to avg_data.pro in IDL SPEDAS.

"""
import logging
import numpy as np
from pyspedas.tplot_tools import store_data, get_data, tnames, time_double


def avg_data(names, trange=None, res=None, width=None,
             newname=None, suffix=None, overwrite=False):
    """
    Get a new tplot variable with averaged data.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    trange: list of float, optional
        Start time, end time.
        If empty, the data start and end time will be used.
    res: float, optional
        Time resolution in seconds for averaging data.
        It can be less than 1 sec.
        Default is 60 sec.
    width: int, optional
        Number of values for the averaging window.
        If res is set, then width is ignored.
    newname: str/list of str, optional
        List of new names for tplot variables.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply.
        Default is '-avg'.
    overwrite: bool, optional
        Replace the existing tplot name.
        Default is False.

    Returns
    -------
    n_names: str/list of str
        List of new pytplot names.

    """

    old_names = tnames(names)

    if res is None and width is None:
        res = 60  # Default is 60 sec

    if names is None or len(old_names) < 1:
        logging.error('avg_data: No valid tplot names were provided.')
        return

    if suffix is None:
        # IDL SPEDAS default suffix is '_avg'
        suffix = '-avg'

    if overwrite:
        n_names = old_names
    elif newname is None:
        n_names = [s + suffix for s in old_names]
    else:
        n_names = newname

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        n_names = [s + suffix for s in old_names]

    for old_idx, old in enumerate(old_names):
        new = n_names[old_idx]

        # Get times and data
        d = get_data(old)
        metadata = get_data(old, metadata=True)

        time = d[0]
        time = np.array(time_double(time))
        time_len = len(time)

        data = np.array(d[1])
        dim = data.shape
        dim0 = dim[0]
        if dim0 != time_len:
            logging.error('avg_data: Data and time length mismatch.')
            continue
        if len(dim) < 2:
            dim1 = 1
        else:
            dim1 = dim[1]

        # Data may also contain v, v1, v2, v3
        process_energies = []
        retain_energies = []
        for i, lend in enumerate(d):
            if i > 1:
                if len(lend) == len(time):
                    process_energies.append(i)
                else:
                    # These will retained in the results as-is
                    retain_energies.append(i)

        process_v = {}
        for i in process_energies:
            process_v[d._fields[i]] = []

        # Find start and end times
        if trange is not None and len(trange) == 2 and trange[0] < trange[1]:
            time_start = time_double(trange[0])
            time_end = time_double(trange[1])
        elif res is not None:
            # Define start and end times similar to IDL
            time_start = np.floor(time[0] / res) * res
            time_end = np.floor(time[-1] / res + 1) * res
        else:
            time_start = time[0]
            time_end = time[-1]

        # Check for empty set
        count_in_range = len(time[(time >= time_start) & (time <= time_end)])
        if time_end <= time_start or count_in_range < 2:
            logging.error('avg_data: No time values in provided time range.')
            continue

        # Find time bins
        time_duration = time_end - time_start

        if res is not None:
            # Given the resolution, compute bins
            dt = res
            bin_count = int(time_duration / dt)
            ind = np.floor((time - time_start) / dt)
        else:
            # Given the width, compute bins
            bins = np.arange(count_in_range)
            ind = np.floor(bins / width)
            bin_count = int(count_in_range / width)
            dt = time_duration / bin_count

        if bin_count < 2:
            msg = 'avg_data: too few bins. Bins=' + str(bin_count) \
                + ', Data points=' + str(count_in_range)
            logging.error(msg)
            continue

        # Split time into bins
        mdt = (time_end - time_start) / dt
        if (mdt - int(mdt) >= 0.5):
            max_ind = np.ceil(mdt)
        else:
            max_ind = np.floor(mdt)
        w1 = np.asarray(ind < 0).nonzero()
        ind[w1] = -1
        w2 = np.asarray(ind >= max_ind).nonzero()
        ind[w2] = -1

        # Find new times
        mx = np.max(ind) + 1
        new_times = (np.arange(mx) + 0.5) * dt + time_start

        # Find new data
        new_data = []
        for i in range(int(max_ind)):
            if i < 0:
                continue

            idx0 = np.asarray(ind == i).nonzero()
            isempty = True if len(idx0) < 1 else False

            if dim1 < 2:
                nd0 = np.nan if isempty else np.nanmean(data[idx0])
            else:
                nd0 = []
                for j in range(dim1):
                    tmp = np.nan if isempty else np.nanmean(data[idx0, j])
                    nd0.append(tmp)

            new_data.append(nd0)

            for ii in process_energies:
                # The following processes v, v1, v2, v3
                dime1 = len(d[ii][0])
                if dime1 < 2:
                    nd1 = np.nan if isempty else np.nanmean(d[ii][idx0])
                else:
                    nd1 = []
                    for j in range(dime1):
                        tmp = np.nan if isempty else np.nanmean(d[ii][idx0, j])
                        nd1.append(tmp)
                process_v[d._fields[ii]].append(nd1)

        # Create the new tplot variable
        data_dict = {'x': new_times, 'y': new_data}
        for i in retain_energies:
            data_dict[d._fields[i]] = d[i]
        for i in process_energies:
            data_dict[d._fields[i]] = process_v[d._fields[i]]

        store_data(new, data=data_dict, attr_dict=metadata)

        logging.info('avg_data was applied to: %s', new)

    return n_names



"""
Convert Cartesian coordinates to spherical coordinates.

This function is similar to cart2spc.pro in IDL SPEDAS.
"""

import numpy as np


def cart2spc(x, y, z):
    """
    Convert Cartesian coordinates to spherical coordinates.

    Parameters:
    -----------
    x : float
        X coordinate in Cartesian system.
    y : float
        Y coordinate in Cartesian system.
    z : float
        Z coordinate in Cartesian system.

    Returns:
    --------
    r : float
        Radial distance from the origin.
    theta : float
        Polar angle (angle from the positive z-axis to the point) in radians. It ranges from [0, pi].
    phi : float
        Azimuthal angle (angle in the xy-plane from the positive x-axis to the point) in radians. It ranges from [0, 2*pi].

    Notes:
    ------
    - Uses the following equations for conversion:
        r = sqrt(x^2 + y^2 + z^2)
        theta = arccos(z/r)
        phi = 2*pi - arccos(x/sqrt(x^2 + y^2)) if y < 0, otherwise arccos(x/sqrt(x^2 + y^2))
    """

    x, y, z = np.array(x), np.array(y), np.array(z)
    r = np.sqrt(x**2 + y**2 + z**2)

    # Calculate phi
    maskn = np.where(y < 0, 1, 0)
    maskp = np.where(y > 0, 1, 0)
    phi = 2*np.pi - maskn * np.arccos(x/np.sqrt(x**2 + y**2)) + maskp * np.arccos(x/np.sqrt(x**2 + y**2))

    # Calculate theta
    theta = np.arccos(z/r)

    return r, theta, phi



from pyspedas.tplot_tools import store_data, options, set_units, set_coords

def erg_pgs_moments_tplot(
    moments,
    x=None,
    prefix='',
    suffix=''
):

    if x is None:
        print('Error, no x-values specified')
        return

    if not isinstance(moments, dict):
        print('Error, the "moments" variable must be a hash table containing the moments')
        return

    moments_tplot_names = []
    for key in moments.keys():
        tplot_name = prefix + '_' + key + suffix
        store_data(tplot_name, data={'x': x, 'y': moments[key]})
        moments_tplot_names.append(tplot_name)

    options(prefix + '_density' + suffix, 'ysubtitle', '[1/cm^3]')
    options(prefix + '_density' + suffix, 'yrange', [1e-3, 5e+1])
    options(prefix + '_density' + suffix, 'ylog', 1)
    set_units(prefix + '_density' + suffix, '1/cm^3')

    options(prefix + '_velocity' + suffix, 'ysubtitle', '[km/s]')
    options(prefix + '_velocity' + suffix, 'yrange', [-2000., 2000.])
    set_units(prefix + '_velocity' + suffix, 'km/s')
    set_coords(prefix + '_velocity' + suffix, 'dsi')

    options(prefix + '_avgtemp' + suffix, 'ysubtitle', '[eV]')
    options(prefix + '_avgtemp' + suffix, 'yrange', [0., 40000.])
    set_units(prefix + '_avgtemp' + suffix, 'eV')

    options(prefix + '_flux' + suffix, 'ysubtitle', '[1/s/cm2]')
    options(prefix + '_flux' + suffix, 'yrange', [-1e9, 1e9])
    set_units(prefix + '_flux' + suffix, '1/(cm^2-sec)')
    set_coords(prefix + '_flux' + suffix, 'DSI')

    options(prefix + '_eflux' + suffix, 'ysubtitle', '[eV/s/cm2]')
    options(prefix + '_eflux' + suffix, 'yrange', [-1e9, 1e9])
    set_units(prefix + '_eflux' + suffix, 'eV/(cm^2-sec)')
    set_coords(prefix + '_eflux' + suffix, 'DSI')

    options(prefix + '_qflux' + suffix, 'ysubtitle', '[eV/s/cm2]')
    options(prefix + '_qflux' + suffix, 'yrange', [-1e12, 1e12])
    set_units(prefix + '_qflux' + suffix, 'eV/(cm^2-sec)')
    set_coords(prefix + '_qflux' + suffix, 'DSI')

    options(prefix + '_t3' + suffix, 'ysubtitle', '[eV]')
    options(prefix + '_t3' + suffix, 'yrange', [-1e12, 1e12])
    set_units(prefix + '_t3' + suffix, 'eV')
    set_coords(prefix + '_t3' + suffix, 'DSI')

    options(prefix + '_magt3' + suffix, 'ysubtitle', '[eV]')
    options(prefix + '_magt3' + suffix, 'yrange', [-1e12, 1e12])
    set_units(prefix + '_magt3' + suffix, 'eV')
    set_coords(prefix + '_magt3' + suffix, 'FA')

    options(prefix + '_mftens' + suffix, 'ysubtitle', '[eV/cm^3]')
    options(prefix + '_mftens' + suffix, 'Color', ['b', 'g', 'r', 'm', 'c', 'y'])
    options(prefix + '_mftens' + suffix, 'legend_names', ['momf_'+label_suffix
            for label_suffix in 'xx yy zz xy xz yz'.split(' ')])
    set_units(prefix + '_mftens' + suffix, 'eV/cm^3')
    set_coords(prefix + '_mftens' + suffix, 'DSI')

    options(prefix + '_ptens' + suffix, 'ysubtitle', '[eV/cm^3]')
    options(prefix + '_ptens' + suffix, 'Color', ['b', 'g', 'r', 'm', 'c', 'y'])
    options(prefix + '_ptens' + suffix, 'legend_names', ['P_'+label_suffix
            for label_suffix in 'xx yy zz xy xz yz'.split(' ')])
    options(prefix + '_ptens' + suffix, 'yrange', [1e+0, 1e+5])
    options(prefix + '_ptens' + suffix, 'ylog', 1)
    set_units(prefix + '_ptens' + suffix, 'eV/cm^3')
    set_coords(prefix + '_ptens' + suffix, 'DSI')

    options(prefix + '_symm_theta' + suffix, 'ysubtitle', '[deg]')
    set_units(prefix + '_symm_theta' + suffix, 'deg')

    options(prefix + '_symm_phi' + suffix, 'ysubtitle', '[deg]')
    set_units(prefix + '_symm_phi' + suffix, 'deg')

    options(prefix + '_symm_ang' + suffix, 'ysubtitle', '[deg]')
    set_units(prefix + '_symm_ang' + suffix, 'deg')

    return moments_tplot_names



import numpy as np

from pyspedas.tplot_tools import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load
from ...satellite.erg.get_gatt_ror import get_gatt_ror


from typing import List, Optional, Union


def gmag_magdas_1sec(
    trange: List[str] = ["2010-11-20/00:00:00", "2010-11-21/00:00:00"],
    suffix: str = "",
    site: Union[str, List[str]] = "all",
    datatype: Union[str, List[str]] = "1sec",
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    Load MAGDAS 1sec data from ERG Science Center

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2010-11-20/00:00:00', '2010-11-21/00:00:00']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load.
            Valid values: "ama", "asb", "daw", "her", "hln", "hob", "kuj", "laq", "mcq", "mgd",
            "mlb", "mut", "onw", "ptk", "wad", "yap", "all"
            Default: ['all']

    datatype: str or list of str
            The data types to load. Valid values: '1s', '1sec', 'all'
            Default: '1sec'

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------

    Examples
    ________
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> magdas_vars=pyspedas.projects.erg.gmag_magdas_1sec(trange=["2010-01-01", "2010-01-02"],site='ama')
    >>> tplot('magdas_mag_ama_1sec_hdz')

    """

    site_code_all = [
        "ama",
        "asb",
        "daw",
        "her",
        "hln",
        "hob",
        "kuj",
        "laq",
        "mcq",
        "mgd",
        "mlb",
        "mut",
        "onw",
        "ptk",
        "wad",
        "yap",
    ]
    tres_all = ["1sec"]
    if isinstance(datatype, str):
        datatype = datatype.lower()
        datatype = datatype.split(" ")
    elif isinstance(datatype, list):
        for i in range(len(datatype)):
            datatype[i] = datatype[i].lower()

    if "all" in datatype:
        datatype = tres_all
    datatype = list(set(datatype).intersection(tres_all))
    if len(datatype) < 1:
        return

    if "1s" in datatype:
        index = np.where(np.array(datatype) == "1s")[0][0]
        datatype[index] = "1sec"

    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(" ")
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if "all" in site_code:
        site_code = site_code_all
    site_code = list(set(site_code).intersection(site_code_all))

    prefix = "magdas_"
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        for data_type_in in datatype:
            fres = data_type_in

            file_res = 3600.0 * 24
            pathformat = (
                "ground/geomag/magdas/"
                + fres
                + "/"
                + site_input
                + "/%Y/magdas_"
                + fres
                + "_"
                + site_input
                + "_%Y%m%d_v??.cdf"
            )

            loaded_data_temp = load(
                pathformat=pathformat,
                file_res=file_res,
                trange=trange,
                datatype=datatype,
                prefix=prefix,
                suffix="_" + site_input + suffix,
                get_support_data=get_support_data,
                varformat=varformat,
                downloadonly=downloadonly,
                notplot=notplot,
                time_clip=time_clip,
                no_update=no_update,
                uname=uname,
                passwd=passwd,
                force_download=force_download,
            )

            if notplot:
                loaded_data.update(loaded_data_temp)
            else:
                loaded_data += loaded_data_temp
            if (len(loaded_data_temp) > 0) and ror:
                try:
                    gatt = get_gatt_ror(downloadonly, loaded_data)
                    print(gatt["Logical_source_description"])
                    print("")
                    print(f'Information about {gatt["Station_code"]}')
                    print(f'PI and Host PI(s): {gatt["PI_name"]}')
                    print("")
                    print("Affiliations: ")
                    print(gatt["PI_affiliation"])
                    print("")
                    print("Rules of the Road for MAGDAS Data Use:")
                    for gatt_text in gatt["TEXT"]:
                        print(gatt_text)
                    print(f'{gatt["LINK_TEXT"]} {gatt["HTTP_LINK"]}')
                    print(
                        "**************************************************************************"
                    )
                except:
                    print("printing PI info and rules of the road was failed")

            if (not downloadonly) and (not notplot):
                current_tplot_name = prefix + "hdz_" + fres + "_" + site_input + suffix
                if current_tplot_name in loaded_data:
                    # ;--- Rename *** HDZ
                    new_tplot_name = (
                        prefix + "mag_" + site_input + "_" + fres + "_hdz" + suffix
                    )
                    store_data(current_tplot_name, newname=new_tplot_name)
                    loaded_data.remove(current_tplot_name)
                    loaded_data.append(new_tplot_name)
                    # ;--- Missing data -1.e+31 --> NaN
                    clip(new_tplot_name, -7e4, 7e4)
                    get_data_vars = get_data(new_tplot_name)
                    ylim(
                        new_tplot_name,
                        -np.nanmax(abs(get_data_vars[1])) * 1.1,
                        np.nanmax(abs(get_data_vars[1])) * 1.1,
                    )
                    # ;--- Labels
                    options(new_tplot_name, "legend_names", ["H", "D", "Z"])
                    options(new_tplot_name, "Color", ["b", "g", "r"])
                    options(
                        new_tplot_name, "ytitle", "\n".join(new_tplot_name.split("_"))
                    )

                current_tplot_name = prefix + "f_" + fres + "_" + site_input + suffix
                if current_tplot_name in loaded_data:
                    # ; --- Rename *** F
                    new_tplot_name = (
                        prefix + "mag_" + site_input + "_" + fres + "_f" + suffix
                    )
                    store_data(current_tplot_name, newname=new_tplot_name)
                    loaded_data.remove(current_tplot_name)
                    loaded_data.append(new_tplot_name)
                    # ;--- Missing data -1.e+31 --> NaN
                    clip(new_tplot_name, -7e4, 7e4)
                    get_data_vars = get_data(new_tplot_name)
                    ylim(
                        new_tplot_name,
                        -np.nanmax(abs(get_data_vars[1])) * 1.1,
                        np.nanmax(abs(get_data_vars[1])) * 1.1,
                    )
                    # ;--- Labels
                    options(new_tplot_name, "legend_names", ["F"])
                    options(
                        new_tplot_name, "ytitle", "\n".join(new_tplot_name.split("_"))
                    )

    return loaded_data




import logging

import numpy as np
from pyspedas import tnames
from pyspedas.tplot_tools import time_double
from pyspedas.tplot_tools import time_string
from pyspedas.tplot_tools import get_data
from scipy import interpolate

from .get_mepe_flux_angle_in_sga import get_mepe_flux_angle_in_sga

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def erg_mepe_get_dist(tname,
                      index=None,
                      units='flux',
                      level='l2',
                      species='e',
                      time_only=False,
                      single_time=None,
                      trange=None):

    if len(tnames(tname)) > 0:
        input_name = tnames(tname)[0]
    else:
        print(f'Variable: {tname} not found!')
        return 0

    level = level.lower()
    """
    ;; Extract some information from a tplot variable name
    ;; e.g., erg_mepe_l2_3dflux_FEDU
    """

    vn_info = input_name.split('_')
    instrument = vn_info[1]
    level = vn_info[2]
    vn_spph = '_'.join(vn_info[0:4]) + '_spin_phase'

    if instrument == 'mepe':
        species = 'e'
    else:
        print(f'ERROR: given an invalid tplot variable: {input_name}')
        return 0

    #  If index is provided, ensure it's a list

    if index is not None and not isinstance(index, list) and not isinstance(index, np.ndarray):
        index = [index]

    # ;; Get a reference to data and metadata

    data_in = get_data(input_name)
    data_in_metadata = get_data(input_name, metadata=True)

    if data_in is None:
        print('Problem extracting the mepe 3dflux data.')
        return 0

    # ;; Return time labels
    if time_only:
        return data_in[0]

    if single_time is not None:
        single_time_double = time_double(single_time)
        if (data_in[0][0] <= single_time_double)\
                and (single_time_double <= data_in[0][-1]):
            nearest_time = interpolate.interp1d(data_in[0], data_in[0],
                                                kind="nearest")(single_time_double)
        elif (single_time_double < data_in[0][0])\
                or (data_in[0][-1] < single_time_double):
            nearest_time = interpolate.interp1d(data_in[0], data_in[0],
                                                kind="nearest", fill_value='extrapolate')(single_time_double)
        index = np.where(data_in[0] == nearest_time)[0]
        n_times = index.size
    else:
        # index supersedes time range
        if index is None:
            if trange is not None:
                trange_double_array = np.array(time_double(trange))
                trange_minmax = np.array([trange_double_array.min(),
                                          trange_double_array.max()])
                index = np.where((data_in[0] >= trange_minmax[0])
                                 & (data_in[0] <= trange_minmax[1]))[0]
                n_times = index.size
                if n_times == 0:
                    print('No data in time range: '
                          + ' '.join(time_string(trange_minmax)))
                    return 0
            else:
                n_times = data_in[0].size
                index = np.arange(n_times)
        else:
            n_times = np.array([index]).size

    """
    ;; --------------------------------------------------------------

    ;; MEPe data arr: [9550(time), 32(spin phase), 16(energy), 16(apd)]
    ;; Dimensions
    """

    # ;; to [ energy, spin phase(azimuth), apd(elevation) ]
    dim_array = np.array(data_in[1].shape[1:])[[1, 0, 2]]

    n_sp = dim_array[1]  # ;; # of spin phases in 1 spin

    if species.lower() == 'e':
        mass = 5.68566e-06
        charge = -1.
        data_name = 'MEP-e Electron 3dflux'
        integ_time = 7.99 / 32 / 16  # ;; currently hard-coded
    else:
        print('given species is not supported by this routine.')
        return 0

    #  ;; basic template structure compatible with other routines

    dist = {
        'project_name': 'ERG',
        'spacecraft': 1,  # always 1 as a dummy value
        'data_name': data_name,
        'units_name': 'flux',  # MEP-e data in [/keV-s-sr-cm2]
                               # should be converted to [/eV-s-sr-cm2]
        'units_procedure': 'erg_convert_flux_units',
        'species': species,
        'valid': 1,

        'charge': charge,
        'mass': mass,

    }

    # other keys (members) will added in below processing, sequentially.

    """
    ;; Then, fill in arrays in the data structure
    ;;   dim[ nenergy, nspinph(azimuth), napd(elevation), ntime]
    """

    dist['time'] = data_in[0][index]
    dist['end_time'] = dist['time'] + integ_time  # ;; currently hard-coded

    """
    ;; Shuffle the original data array [time,spin phase,energy,apd] to
    ;; be energy-azimuth-elevation-time.
    ;; The factor 1d-3 is to convert [/keV-s-sr-cm2] (default unit of
    ;; MEP-e Lv2 flux data) to [/eV-s-sr-cm2]
    """
    dist['data'] = data_in[1][index].transpose([2, 1, 3, 0]) * 1e-3

    dist['bins'] = np.ones(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), dtype='int8')
    # must be set or data will be consider invalid

    """
    ;; Energy ch. 0 is excluded due to difficulty in defining
    ;; the representative energy and energy bin width.
    """
    dist['bins'][0] = 0

    #  ;; Energy ch
    e0_array = data_in[3] * 1e+3  # ;; [keV] 
                            #(default of MEP-e Lv2 flux data) to [eV]
    energy_reform = np.reshape(e0_array, [dim_array[0], 1, 1, 1])
    energy_rebin1 = np.repeat(energy_reform, dim_array[2],
                             axis=2)  # repeated across apd(elevation)
    energy_rebin2 = np.repeat(energy_rebin1, dim_array[1],
                              axis=1)  # repeated across spin phase(azimuth)
    dist['energy'] = np.repeat(energy_rebin2, n_times,
                              axis=3)  # repeated across n_times

    #  ;; Energy bin width
    e0bnd = np.sqrt(e0_array[:-1] * e0_array[1:])
    # ;; [16]  upper boundary of energy bin
    e0bnd_p = np.insert(e0bnd, 0, e0bnd[0])
    e0bnd_p[[0, 1]] = e0_array[1] + (e0_array[1]-e0bnd[1])

    # ;; [16] lower boundary of energy bin
    e0bnd_m = np.insert(e0bnd, e0bnd.shape[0], e0bnd[-1])
    e0bnd_m[0] = e0bnd_m[1]
    e0bnd_m[15] = e0_array[15] - (e0bnd[14]-e0_array[15])

    de_array = e0bnd_p-e0bnd_m  # ;; width of energy bin
    de_reform = np.reshape(de_array, [dim_array[0], 1, 1, 1])
    de_rebin1 = np.repeat(de_reform, dim_array[2],
                         axis=2)  # repeated across apd(elevation)
    de_rebin2 = np.repeat(de_rebin1, dim_array[1],
                         axis=1)  # repeated across spin phase(azimuth)
    dist['denergy'] = np.repeat(de_rebin2, n_times,
                         axis=3)  # repeated across n_times

    dist['n_energy'] = dim_array[0]

    dist['n_bins'] = dim_array[1] * dim_array[2]  #   # thetas * # phis
   
    #  ;; azimuthal angle in spin direction

    angarr = get_mepe_flux_angle_in_sga()
    phissi = angarr[1, 1, :] - (90. + 21.6)  # ;; [(apd)]
    spinph_ofst = data_in[2] * 11.25

    phi0_1_reform = np.reshape(phissi, [1, 1, dim_array[2], 1])
    phi0_1_rebin1 = np.repeat(phi0_1_reform, dim_array[1],
                             axis=1)  # repeated across spin phase(azimuth)

    phi0_1_rebin2 = np.repeat(phi0_1_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    phi0_1 = np.repeat(phi0_1_rebin2,     n_times,
                              axis=3)  # repeated across n_times
    phi0_2_reform = np.reshape(spinph_ofst, [1, dim_array[1], 1, 1])
    phi0_2_rebin1 = np.repeat(phi0_2_reform, dim_array[2],
                             axis=2)  # repeated across apd(elevation)
   
    phi0_2_rebin2 = np.repeat(phi0_2_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    phi0_2 = np.repeat(phi0_2_rebin2, n_times,
                              axis=3)  # repeated across n_times
    phi0 = phi0_1 + phi0_2

    ofst_sv = (np.arange(dim_array[0]) + 0.5) * \
        11.25/dim_array[0]  # ;; [(energy)]
    phi_ofst_for_sv_reform = np.reshape(ofst_sv, [dim_array[0], 1, 1, 1])
    phi_ofst_for_sv_rebin1 = np.repeat(
        phi_ofst_for_sv_reform, dim_array[2],
                                 axis=2)  # repeated across apd(elevation)

    phi_ofst_for_sv_rebin2 = np.repeat(
        phi_ofst_for_sv_rebin1,dim_array[1],
                                 axis=1)  # repeated across spin phase(azimuth)
    phi_ofst_for_sv = np.repeat(phi_ofst_for_sv_rebin2,         n_times,
                                 axis=3)  # repeated across n_times
    """
    ;;  phi angle for the start of each spin phase
    ;;    + offset angle foreach sv step
    """
    dist['phi'] = np.fmod((phi0 + phi_ofst_for_sv + 360.), 360.)
    dist['dphi'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), fill_value=11.25)

    del phi0, phi_ofst_for_sv  # ;; Clean huge arrays

    dist['n_phi'] = dim_array[1]

    #  ;; elevation angle
    elev = angarr[0, 1, :]  # ;; [(apd)]
    elev_reform = np.reshape(elev, [1, 1, dim_array[2], 1])
    elev_rebin1 = np.repeat(elev_reform, dim_array[1],
                             axis=1)  # repeated across spin phase(azimuth)
    elev_rebin2 = np.repeat(elev_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    dist['theta'] = np.repeat(elev_rebin2, n_times,
                             axis=3)  # repeated across n_times

    dist['dtheta'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                             n_times), fill_value=11.25)  # ;; 11.25 deg is set for the moment calculation

    dist['n_theta'] = dim_array[2]

    return dist


import numpy as np

from pyspedas.tplot_tools import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load
from ...satellite.erg.get_gatt_ror import get_gatt_ror


from typing import List, Optional, Union


def gmag_mm210(
    trange: List[str] = ["2020-08-01", "2020-08-02"],
    suffix: str = "",
    site: Union[str, List[str]] = "all",
    datatype: Union[str, List[str]] = "all",
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    Load mm210 data from ERG Science Center

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2020-08-01', '2020-08-02']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load.
            Valid values:

                    'tik', 'zgn', 'yak', 'irt', 'ppi', 'bji',
                    'lnp', 'mut', 'ptn', 'wtk', 'lmt', 'kat',
                    'ktn', 'chd', 'zyk', 'mgd', 'ptk', 'msr',
                    'rik', 'onw', 'kag', 'ymk', 'cbi', 'gua',
                    'yap', 'kor', 'ktb', 'bik', 'wew', 'daw',
                    'wep', 'bsv', 'dal', 'can', 'adl', 'kot',
                    'cst', 'ewa', 'asa', 'mcq', 'all'

            Default: ['all']

    datatype: str or list of str
            The data types to load. Valid values: '1s', '1sec', '1m', '1min', '1h', '1hr', 'all'
            Default: 'all'

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------

    Examples
    ________
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> mm210_vars=pyspedas.projects.erg.gmag_mm210(trange=["2005-01-01", "2005-01-02"],site='adl',datatype='1min')
    >>> tplot('mm210_mag_adl_1min_hdz')

    """

    site_code_all = [
        "tik",
        "zgn",
        "yak",
        "irt",
        "ppi",
        "bji",
        "lnp",
        "mut",
        "ptn",
        "wtk",
        "lmt",
        "kat",
        "ktn",
        "chd",
        "zyk",
        "mgd",
        "ptk",
        "msr",
        "rik",
        "onw",
        "kag",
        "ymk",
        "cbi",
        "gua",
        "yap",
        "kor",
        "ktb",
        "bik",
        "wew",
        "daw",
        "wep",
        "bsv",
        "dal",
        "can",
        "adl",
        "kot",
        "cst",
        "ewa",
        "asa",
        "mcq",
    ]
    tres_all = ["1sec", "1min", "1h"]
    if isinstance(datatype, str):
        datatype = datatype.lower()
        datatype = datatype.split(" ")
    elif isinstance(datatype, list):
        for i in range(len(datatype)):
            datatype[i] = datatype[i].lower()

    if "all" in datatype:
        datatype = tres_all
    datatype = list(set(datatype).intersection(tres_all))
    if len(datatype) < 1:
        return

    if "1s" in datatype:
        index = np.where(np.array(datatype) == "1s")[0][0]
        datatype[index] = "1sec"
    elif "1m" in datatype:
        index = np.where(np.array(datatype) == "1m")[0][0]
        datatype[index] = "1min"
    elif "1hr" in datatype:
        index = np.where(np.array(datatype) == "1hr")[0][0]
        datatype[index] = "1h"

    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(" ")
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if "all" in site_code:
        site_code = site_code_all

    site_code = list(set(site_code).intersection(site_code_all))

    prefix = "mm210_"
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        for data_type_in in datatype:
            fres = data_type_in

            if fres == "1h":
                fres = "1min"

            file_res = 3600.0 * 24
            pathformat = (
                "ground/geomag/mm210/"
                + fres
                + "/"
                + site_input
                + "/%Y/mm210_"
                + fres
                + "_"
                + site_input
                + "_%Y%m%d_v??.cdf"
            )

            loaded_data_temp = load(
                pathformat=pathformat,
                file_res=file_res,
                trange=trange,
                datatype=datatype,
                prefix=prefix,
                suffix="_" + site_input + suffix,
                get_support_data=get_support_data,
                varformat=varformat,
                downloadonly=downloadonly,
                notplot=notplot,
                time_clip=time_clip,
                no_update=no_update,
                uname=uname,
                passwd=passwd,
                force_download=force_download,
            )

            if notplot:
                loaded_data.update(loaded_data_temp)
            else:
                loaded_data += loaded_data_temp
            if (len(loaded_data_temp) > 0) and ror:
                try:
                    gatt = get_gatt_ror(downloadonly, loaded_data)
                    print(
                        "**************************************************************************"
                    )
                    print(gatt["Logical_source_description"])
                    print("")
                    print(f'Information about {gatt["Station_code"]}')
                    print("PI and Host PI(s):")
                    print(gatt["PI_name"])
                    print("")
                    print("Affiliations: ")
                    print(gatt["PI_affiliation"])
                    print("")
                    print("Rules of the Road for 210 MM Data Use:")
                    for gatt_text in gatt["TEXT"]:
                        print(gatt_text)
                    print(f'{gatt["LINK_TEXT"]} {gatt["HTTP_LINK"]}')
                    print(
                        "**************************************************************************"
                    )
                except:
                    print("printing PI info and rules of the road was failed")

            if (not downloadonly) and (not notplot):
                if fres == "1min":
                    fres_list = ["1min", "1h"]
                else:
                    fres_list = [fres]
                for fres_in in fres_list:
                    current_tplot_name = (
                        prefix + "hdz_" + fres_in + "_" + site_input + suffix
                    )
                    if current_tplot_name in loaded_data:
                        get_data_vars = get_data(current_tplot_name)
                        if get_data_vars is None:
                            store_data(current_tplot_name, delete=True)
                        else:
                            # ;--- Rename
                            new_tplot_name = (
                                prefix
                                + "mag_"
                                + site_input
                                + "_"
                                + fres_in
                                + "_hdz"
                                + suffix
                            )
                            store_data(current_tplot_name, newname=new_tplot_name)
                            loaded_data.remove(current_tplot_name)
                            loaded_data.append(new_tplot_name)
                            # ;--- Missing data -1.e+31 --> NaN
                            clip(new_tplot_name, -1e4, 1e4)
                            get_data_vars = get_data(new_tplot_name)
                            ylim(
                                new_tplot_name,
                                np.nanmin(get_data_vars[1]),
                                np.nanmax(get_data_vars[1]),
                            )
                            # ;--- Labels
                            options(
                                new_tplot_name, "legend_names", ["Ch1", "Ch2", "Ch3"]
                            )
                            options(new_tplot_name, "Color", ["b", "g", "r"])
                            options(
                                new_tplot_name,
                                "ytitle",
                                "\n".join(new_tplot_name.split("_")),
                            )

    return loaded_data
